# these tests are run using the run-in-pyodide option from test_run_in_pyodide.py
# test_fail is expected to fail obviously!
import sys


def test_success():
    print("WOOO")


def test_fail():
    print("Oh dear")
    sys.stderr.write("Oops, this is gonna fail")
    assert 1 == 0


def test_check_in_pyodide():
    print("Check in pyodide")
    try:
        import pyodide

        dir(pyodide)
    except ImportError:
        assert "Not in pyodide" == 0


def test_this_doesnt_run(selenium):
    assert "Pyodide specific tests shouldn't be run inside pyodide" == 0



import json
import textwrap
from pathlib import Path

import pexpect
import pytest

CHROME_FLAGS: list[str] = ["--js-flags=--expose-gc"]
FIREFOX_FLAGS: list[str] = []
NODE_FLAGS: list[str] = []


TEST_SETUP_CODE = """
Error.stackTraceLimit = Infinity;

// Fix globalThis is messed up in firefox see facebook/react#16606.
// Replace it with window.
globalThis.globalThis = globalThis.window || globalThis;

globalThis.sleep = function (s) {
    return new Promise((resolve) => setTimeout(resolve, s));
};

globalThis.assert = function (cb, message = "") {
    if (message !== "") {
        message = "\\n" + message;
    }
    if (cb() !== true) {
        throw new Error(
            `Assertion failed: ${cb.toString().slice(6)}${message}`
        );
    }
};

globalThis.assertAsync = async function (cb, message = "") {
    if (message !== "") {
        message = "\\n" + message;
    }
    if ((await cb()) !== true) {
        throw new Error(
            `Assertion failed: ${cb.toString().slice(12)}${message}`
        );
    }
};

function checkError(err, errname, pattern, pat_str, thiscallstr) {
    if (typeof pattern === "string") {
        pattern = new RegExp(pattern);
    }
    if (!err) {
        throw new Error(`${thiscallstr} failed, no error thrown`);
    }
    if (err.constructor.name !== errname) {
        throw new Error(
            `${thiscallstr} failed, expected error ` +
                `of type '${errname}' got type '${err.constructor.name}'`
        );
    }
    if (!pattern.test(err.message)) {
        throw new Error(
            `${thiscallstr} failed, expected error ` +
                `message to match pattern ${pat_str} got:\n${err.message}`
        );
    }
}

globalThis.assertThrows = function (cb, errname, pattern) {
    let pat_str = typeof pattern === "string" ? `"${pattern}"` : `${pattern}`;
    let thiscallstr = `assertThrows(${cb.toString()}, "${errname}", ${pat_str})`;
    let err = undefined;
    try {
        cb();
    } catch (e) {
        err = e;
    }
    checkError(err, errname, pattern, pat_str, thiscallstr);
};

globalThis.assertThrowsAsync = async function (cb, errname, pattern) {
    let pat_str = typeof pattern === "string" ? `"${pattern}"` : `${pattern}`;
    let thiscallstr = `assertThrowsAsync(${cb.toString()}, "${errname}", ${pat_str})`;
    let err = undefined;
    try {
        await cb();
    } catch (e) {
        err = e;
    }
    checkError(err, errname, pattern, pat_str, thiscallstr);
};
""".strip()

INITIALIZE_SCRIPT = "pyodide.runPython('');"


class JavascriptException(Exception):
    def __init__(self, msg, stack):
        self.msg = msg
        self.stack = stack
        # In chrome the stack contains the message
        if self.stack and self.stack.startswith(self.msg):
            self.msg = ""

    def __str__(self):
        return "\n\n".join(x for x in [self.msg, self.stack] if x)


class _BrowserBaseRunner:
    browser = ""
    script_timeout = 20
    JavascriptException = JavascriptException

    def __init__(
        self,
        server_port,
        server_hostname="127.0.0.1",
        server_log=None,
        load_pyodide=True,
        script_type="classic",
        dist_dir=None,
        jspi=False,
        *args,
        **kwargs,
    ):
        self.server_port = server_port
        self.server_hostname = server_hostname
        self.base_url = f"http://{self.server_hostname}:{self.server_port}"
        self.server_log = server_log
        self.script_type = script_type
        self.dist_dir = dist_dir
        self.driver = self.get_driver(jspi)
        self.set_script_timeout(self.script_timeout)
        self.prepare_driver()
        self.javascript_setup()
        if load_pyodide:
            self.load_pyodide()
            self.initialize_pyodide()
            self.save_state()
            self.restore_state()

    def get_driver(self, jspi=False):
        raise NotImplementedError()

    def goto(self, page):
        raise NotImplementedError()

    def set_script_timeout(self, timeout):
        raise NotImplementedError()

    def quit(self):
        raise NotImplementedError()

    def refresh(self):
        raise NotImplementedError()

    def run_js_inner(self, code, check_code):
        raise NotImplementedError()

    def prepare_driver(self):
        if self.script_type == "classic":
            self.goto(f"{self.base_url}/test.html")
        elif self.script_type == "module":
            self.goto(f"{self.base_url}/module_test.html")
        else:
            raise Exception("Unknown script type to load!")

    def javascript_setup(self):
        self.run_js(
            TEST_SETUP_CODE,
            pyodide_checks=False,
        )

    def load_pyodide(self):
        self.run_js(
            """
            let pyodide = await loadPyodide({ fullStdLib: false, jsglobals : self });
            self.pyodide = pyodide;
            globalThis.pyodide = pyodide;
            pyodide._api.inTestHoist = true; // improve some error messages for tests
            """
        )

    def initialize_pyodide(self):
        self.run_js(
            """
            let isPyProxy;
            if(pyodide.ffi) {
                isPyProxy = (o) => o instanceof pyodide.ffi.PyProxy;
            } else {
                isPyProxy = pyodide.isPyProxy;
            }
            pyodide.$handleTestResult = function(result) {
                if(!(result && result.toJs)){
                    return result;
                }
                let converted_result = result.toJs();
                if(isPyProxy(converted_result)){
                    converted_result = undefined;
                }
                result.destroy();
                return converted_result;
            }
            """
        )
        self.run_js(INITIALIZE_SCRIPT)
        from .decorator import initialize_decorator

        initialize_decorator(self)

    @property
    def pyodide_loaded(self):
        return self.run_js("return !!(self.pyodide && self.pyodide.runPython);")

    @property
    def logs(self):
        logs = self.run_js("return self.logs;", pyodide_checks=False)
        if logs is not None:
            return "\n".join(str(x) for x in logs)
        return ""

    def clean_logs(self):
        self.run_js("self.logs = []", pyodide_checks=False)

    def run(self, code):
        return self.run_js(
            f"""
            let result = pyodide.runPython({code!r});
            return pyodide.$handleTestResult(result);
            """
        )

    def run_async(self, code):
        return self.run_js(
            f"""
            await pyodide.loadPackagesFromImports({code!r})
            let result = await pyodide.runPythonAsync({code!r});
            return pyodide.$handleTestResult(result);
            """
        )

    def run_js(self, code, pyodide_checks=True):
        """Run JavaScript code and check for pyodide errors"""
        if isinstance(code, str) and code.startswith("\n"):
            # we have a multiline string, fix indentation
            code = textwrap.dedent(code)

        if pyodide_checks:
            check_code = """
                    if(globalThis.pyodide && pyodide._module && pyodide._module._PyErr_Occurred()){
                        try {
                            pyodide._module._pythonexc2js();
                        } catch(e){
                            console.error(`Python exited with error flag set! Error was:\n${e.message}`);
                            // Don't put original error message in new one: we want
                            // "pytest.raises(xxx, match=msg)" to fail
                            throw new Error(`Python exited with error flag set!`);
                        }
                    }
           """
        else:
            check_code = ""
        return self.run_js_inner(code, check_code)

    def get_num_hiwire_keys(self):
        return self.run_js("return pyodide._module.hiwire.num_keys();")

    @property
    def force_test_fail(self) -> bool:
        return self.run_js("return !!pyodide._api.fail_test;")

    def clear_force_test_fail(self):
        self.run_js("pyodide._api.fail_test = false;")

    def save_state(self):
        self.run_js("self.__savedState = pyodide._api.saveState();")

    def restore_state(self):
        self.run_js(
            """
            if(self.__savedState){
                pyodide._api.restoreState(self.__savedState)
            }
            """
        )

    def get_num_proxies(self):
        return self.run_js("return pyodide._module.pyproxy_alloc_map.size")

    def enable_pyproxy_tracing(self):
        self.run_js("pyodide._module.enable_pyproxy_allocation_tracing()")

    def disable_pyproxy_tracing(self):
        self.run_js("pyodide._module.disable_pyproxy_allocation_tracing()")

    def run_webworker(self, code):
        if isinstance(code, str) and code.startswith("\n"):
            # we have a multiline string, fix indentation
            code = textwrap.dedent(code)

        worker_file = (
            "webworker_dev.js"
            if self.script_type == "classic"
            else "module_webworker_dev.js"
        )

        return self.run_js(
            """
            let worker = new Worker('{}', {{ type: '{}' }});
            let res = new Promise((res, rej) => {{
                worker.onerror = e => rej(e);
                worker.onmessage = e => {{
                    if (e.data.results) {{
                       res(e.data.results);
                    }} else {{
                       rej(e.data.error);
                    }}
                }};
                worker.postMessage({{ python: {!r} }});
            }});
            return await res
            """.format(
                f"http://{self.server_hostname}:{self.server_port}/{worker_file}",
                self.script_type,
                code,
            ),
            pyodide_checks=False,
        )

    def load_package(self, packages):
        self.run_js(f"await pyodide.loadPackage({packages!r})")


class _SeleniumBaseRunner(_BrowserBaseRunner):
    def goto(self, page):
        self.driver.get(page)

    def set_script_timeout(self, timeout):
        self.driver.set_script_timeout(timeout)
        self.script_timeout = timeout

    def quit(self):
        self.driver.quit()

    def refresh(self):
        self.driver.refresh()
        self.javascript_setup()

    def run_js_inner(self, code, check_code):
        wrapper = """
            let cb = arguments[arguments.length - 1];
            let run = async () => { %s }
            (async () => {
                try {
                    let result = await run();
                    %s
                    cb([0, result]);
                } catch (e) {
                    cb([1, e.toString(), e.stack, e.message]);
                }
            })()
        """
        retval = self.driver.execute_async_script(wrapper % (code, check_code))
        if retval[0] == 0:
            return retval[1]
        else:
            print("JavascriptException message: ", retval[3])
            raise JavascriptException(retval[1], retval[2])

    @property
    def urls(self):
        for handle in self.driver.window_handles:
            self.driver.switch_to.window(handle)
            yield self.driver.current_url


class _PlaywrightBaseRunner(_BrowserBaseRunner):
    def __init__(self, browsers, *args, **kwargs):
        self.browsers = browsers
        super().__init__(*args, **kwargs)

    def goto(self, page):
        self.driver.goto(page)

    def get_driver(self, jspi=False):
        if jspi:
            raise NotImplementedError("JSPI not supported with playwright")
        return self.browsers[self.browser].new_page()

    def set_script_timeout(self, timeout):
        # playwright uses milliseconds for timeout
        self.driver.set_default_timeout(timeout * 1000)

    def quit(self):
        self.driver.close()

    def refresh(self):
        self.driver.reload()
        self.javascript_setup()

    def run_js_inner(self, code, check_code):
        # playwright `evaluate` waits until primise to resolve,
        # so we don't need to use a callback like selenium.
        wrapper = """
            let run = async () => { %s }
            (async () => {
                try {
                    let result = await run();
                    %s
                    return [0, result];
                } catch (e) {
                    return [1, e.toString(), e.stack];
                }
            })()
        """
        retval = self.driver.evaluate(wrapper % (code, check_code))
        if retval[0] == 0:
            return retval[1]
        else:
            raise JavascriptException(retval[1], retval[2])


class SeleniumFirefoxRunner(_SeleniumBaseRunner):
    browser = "firefox"

    def get_driver(self, jspi=False):
        if jspi:
            raise NotImplementedError("JSPI not supported in Firefox")
        from selenium.webdriver import Firefox
        from selenium.webdriver.firefox.options import Options
        from selenium.webdriver.firefox.service import Service

        options = Options()
        options.add_argument("--headless")
        for flag in FIREFOX_FLAGS:
            options.add_argument(flag)

        return Firefox(service=Service(), options=options)


class SeleniumChromeRunner(_SeleniumBaseRunner):
    browser = "chrome"

    def get_driver(self, jspi=False):
        from selenium.webdriver import Chrome
        from selenium.webdriver.chrome.options import Options

        options = Options()
        options.add_argument("--headless")
        options.add_argument("--no-sandbox")
        if jspi:
            options.add_argument("--enable-features=WebAssemblyExperimentalJSPI")
            options.add_argument("--enable-experimental-webassembly-features")
        for flag in CHROME_FLAGS:
            options.add_argument(flag)
        return Chrome(options=options)

    def collect_garbage(self):
        self.driver.execute_cdp_cmd("HeapProfiler.collectGarbage", {})


# Stopping and starting the safari webdriver multiple times during the test
# often causes unexpected errors. So we make a global webdriver instance and
# reuse it.
GLOBAL_SAFARI_WEBDRIVER = None


@pytest.fixture(scope="session", autouse=True)
def use_global_safari_service():
    if "safari" in pytest.pyodide_runtimes:  # type: ignore[operator]
        global GLOBAL_SAFARI_WEBDRIVER

        from selenium.webdriver.common.driver_finder import DriverFinder
        from selenium.webdriver.safari.options import Options
        from selenium.webdriver.safari.service import Service

        GLOBAL_SAFARI_WEBDRIVER = Service(reuse_service=True)
        GLOBAL_SAFARI_WEBDRIVER.path = DriverFinder.get_path(
            GLOBAL_SAFARI_WEBDRIVER, Options()
        )
        GLOBAL_SAFARI_WEBDRIVER.start()

        try:
            yield GLOBAL_SAFARI_WEBDRIVER
        finally:
            GLOBAL_SAFARI_WEBDRIVER.stop()
    else:
        yield None


class SeleniumSafariRunner(_SeleniumBaseRunner):
    browser = "safari"
    script_timeout = 30

    def get_driver(self, jspi=False):
        if jspi:
            raise NotImplementedError("JSPI not supported in Firefox")
        from selenium.webdriver import Safari
        from selenium.webdriver.safari.options import Options

        options = Options()
        if GLOBAL_SAFARI_WEBDRIVER is not None:
            instance = Safari(
                options=options, service=GLOBAL_SAFARI_WEBDRIVER, reuse_service=True
            )
        else:
            instance = Safari(options=options)

        return instance


class PlaywrightChromeRunner(_PlaywrightBaseRunner):
    browser = "chrome"

    def collect_garbage(self):
        client = self.driver.context.new_cdp_session(self.driver)
        client.send("HeapProfiler.collectGarbage")


class PlaywrightFirefoxRunner(_PlaywrightBaseRunner):
    browser = "firefox"


class NodeRunner(_BrowserBaseRunner):
    browser = "node"

    def init_node(self, jspi=False):
        curdir = Path(__file__).parent
        self.p = pexpect.spawn("/bin/bash", timeout=60)
        self.p.setecho(False)
        self.p.delaybeforesend = None
        # disable canonical input processing mode to allow sending longer lines
        # See: https://pexpect.readthedocs.io/en/stable/api/pexpect.html#pexpect.spawn.send
        self.p.sendline("stty -icanon")

        node_version = pexpect.spawn("node --version").read().decode("utf-8")
        node_major = int(node_version.split(".")[0][1:])  # vAA.BB.CC -> AA
        if node_major < 18:
            raise RuntimeError(
                f"Node version {node_version} is too old, please use node >= 18"
            )

        extra_args = NODE_FLAGS[:]
        # Node v14 require the --experimental-wasm-bigint which
        # produces errors on later versions
        if jspi:
            extra_args.append("--experimental-wasm-stack-switching")

        self.p.sendline(
            f"node --expose-gc {' '.join(extra_args)} {curdir}/node_test_driver.js {self.base_url} {self.dist_dir}",
        )

        try:
            self.p.expect_exact("READY!!")
        except (pexpect.exceptions.EOF, pexpect.exceptions.TIMEOUT):
            raise JavascriptException("", self.p.before.decode()) from None

    def get_driver(self, jspi=False):
        self._logs = []
        self.init_node(jspi)

        class NodeDriver:
            def __getattr__(self, x):
                raise NotImplementedError()

        return NodeDriver()

    def prepare_driver(self):
        pass

    def set_script_timeout(self, timeout):
        self.script_timeout = timeout

    def quit(self):
        self.p.sendeof()

    def refresh(self):
        self.quit()
        self.init_node()
        self.javascript_setup()

    def collect_garbage(self):
        self.run_js("gc()")

    @property
    def logs(self):
        return "\n".join(self._logs)

    def clean_logs(self):
        self._logs = []

    def run_js_inner(self, code, check_code):
        check_code = ""
        wrapped = """
            let result = await (async () => {{ {} }})();
            {}
            return result;
        """.format(
            code,
            check_code,
        )
        from uuid import uuid4

        cmd_id = str(uuid4())
        self.p.sendline(cmd_id)
        self.p.sendline(wrapped)
        self.p.sendline(cmd_id)
        self.p.expect_exact(f"{cmd_id}:UUID\r\n", timeout=self.script_timeout)
        self.p.expect_exact(f"{cmd_id}:UUID\r\n")
        if self.p.before:
            self._logs.append(self.p.before.decode()[:-2].replace("\r", ""))
        self.p.expect("[01]\r\n")
        success = int(self.p.match[0].decode()[0]) == 0
        self.p.expect_exact(f"\r\n{cmd_id}:UUID\r\n")
        if success:
            return json.loads(self.p.before.decode().replace("undefined", "null"))
        else:
            raise JavascriptException("", self.p.before.decode())



import contextlib
import functools
import http.server
import multiprocessing
import os
import pathlib
import queue
import shutil
import socketserver
import sys
import tempfile
from io import BytesIO


@functools.cache
def _default_templates() -> dict[str, bytes]:
    templates_dir = pathlib.Path(__file__).parent / "_templates"

    templates = {}
    for template_file in templates_dir.glob("*.html"):
        templates[f"/{template_file.name}"] = template_file.read_bytes()

    return templates


class DefaultHandler(http.server.SimpleHTTPRequestHandler):
    default_templates = _default_templates()

    def __init__(self, *args, **kwargs):
        self.extra_headers = kwargs.pop("extra_headers", {})
        super().__init__(*args, **kwargs)

    def log_message(self, format_, *args):
        print(
            "[%s] source: %s:%s - %s"
            % (
                self.log_date_time_string(),
                *self.client_address,
                format_ % args,
            )
        )

    def get_template(self, path: str) -> bytes | None:
        """
        Return the content of the template if it exists, None otherwise

        This method is used to serve the default templates, and can be
        overridden to serve custom templates.
        """
        return self.default_templates.get(path)

    def do_GET(self):
        body = self.get_template(self.path)
        if body:
            self.send_response(200)
            self.send_header("Content-type", "text/html; charset=utf-8")
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()

            self.copyfile(BytesIO(body), self.wfile)
        else:
            return super().do_GET()

    def end_headers(self):
        # Enable Cross-Origin Resource Sharing (CORS)
        self.send_header("Access-Control-Allow-Origin", "*")
        for k, v in self.extra_headers.items():
            self.send_header(k, v)
        if len(self.extra_headers) > 0:
            joined_headers = ",".join(self.extra_headers.keys())
            # if you don't send this, CORS blocks custom headers in javascript
            self.send_header("Access-Control-Expose-Headers", joined_headers)
        super().end_headers()


@contextlib.contextmanager
def spawn_web_server(dist_dir, extra_headers=None, handler_cls=None):
    if not extra_headers:
        extra_headers = {}
    tmp_dir = tempfile.mkdtemp()
    log_path = pathlib.Path(tmp_dir) / "http-server.log"
    q: multiprocessing.Queue[str] = multiprocessing.Queue()
    p = multiprocessing.Process(
        target=run_web_server, args=(q, log_path, dist_dir, extra_headers, handler_cls)
    )

    try:
        p.start()
        port = q.get(timeout=20)
        hostname = "127.0.0.1"

        print(
            f"Spawning webserver at http://{hostname}:{port} "
            f"(see logs in {log_path})"
        )
        yield hostname, port, log_path
    finally:
        q.put("TERMINATE")
        p.join()
        shutil.rmtree(tmp_dir)


def run_web_server(q, log_filepath, dist_dir, extra_headers, handler_cls):
    """Start the HTTP web server

    Parameters
    ----------
    q : Queue
      communication queue
    log_path : pathlib.Path
      path to the file where to store the logs
    """

    os.chdir(dist_dir)

    log_fh = log_filepath.open("w", buffering=1)
    sys.stdout = log_fh
    sys.stderr = log_fh

    if not handler_cls:
        handler_cls = functools.partial(DefaultHandler, extra_headers=extra_headers)

    with socketserver.TCPServer(("", 0), handler_cls) as httpd:
        host, port = httpd.server_address
        print(f"Starting webserver at http://{host}:{port}")  # type: ignore[str-bytes-safe]
        httpd.server_name = "test-server"  # type: ignore[attr-defined]
        httpd.server_port = port  # type: ignore[attr-defined]
        q.put(port)

        def service_actions():
            try:
                if q.get(False) == "TERMINATE":
                    print("Stopping server...")
                    sys.exit(0)
            except queue.Empty:
                pass

        httpd.service_actions = service_actions  # type: ignore[method-assign]
        httpd.serve_forever()



import http.server
from http import HTTPStatus

import requests

from pytest_pyodide.server import DefaultHandler, _default_templates, spawn_web_server


def test_spawn_web_server_with_params(tmp_path):
    (tmp_path / "index.txt").write_text("a")
    extra_headers = {"Custom-Header": "42"}
    with spawn_web_server(tmp_path, extra_headers=extra_headers) as (
        hostname,
        port,
        log_path,
    ):
        res = requests.get(f"http://{hostname}:{port}/index.txt")
        assert res.ok
        assert res.headers
        assert res.content == b"a"
        assert res.headers["Access-Control-Allow-Origin"] == "*"
        assert res.headers.get("Custom-Header") == "42"


def test_spawn_web_server_default_templates(tmp_path):
    default_templates = _default_templates()

    with spawn_web_server(tmp_path) as (hostname, port, _):
        for path, content in default_templates.items():
            res = requests.get(f"http://{hostname}:{port}{path}")
            assert res.ok
            assert res.headers
            assert res.content == content
            assert res.headers["Access-Control-Allow-Origin"] == "*"


class CustomTemplateHandler(DefaultHandler):
    def get_template(self, path: str) -> bytes | None:
        if path == "/index.txt":
            return b"hello world"

        return super().get_template(path)


def test_spawn_web_server_custom_templates(tmp_path):
    default_templates = _default_templates()

    with spawn_web_server(tmp_path, handler_cls=CustomTemplateHandler) as (
        hostname,
        port,
        _,
    ):
        for path, content in default_templates.items():
            res = requests.get(f"http://{hostname}:{port}{path}")
            assert res.ok
            assert res.headers
            assert res.content == content
            assert res.headers["Access-Control-Allow-Origin"] == "*"

        res = requests.get(f"http://{hostname}:{port}/index.txt")
        assert res.ok
        assert res.content == b"hello world"


class HelloWorldHandler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        self.send_response(HTTPStatus.OK)
        self.end_headers()
        self.wfile.write(b"hello world")


def test_custom_handler(tmp_path):
    with spawn_web_server(tmp_path, handler_cls=HelloWorldHandler) as server:
        hostname, port, _ = server
        res = requests.get(f"http://{hostname}:{port}/index.txt")
        assert res.ok
        assert res.content == b"hello world"



from pathlib import Path
from textwrap import dedent

import pytest

from pytest_pyodide import run_in_pyodide

DOCTESTS = """\
def pyodide_fail():
    '''
    >>> from js import Object # doctest: +RUN_IN_PYODIDE
    >>> 1 == 2
    True
    '''

def host_success():
    '''
    >>> import sys
    >>> sys.platform == "emscripten"
    False
    '''

def pyodide_success():
    '''
    >>> pyodide_success() # doctest: +RUN_IN_PYODIDE
    7
    >>> from js import Object
    >>> import sys
    >>> sys.platform == "emscripten"
    True
    '''
    return 7
"""


def test_doctest_run(pytester, selenium, request, playwright_browsers, capsys):
    file = pytester.makepyfile(DOCTESTS)
    config = pytester.parseconfigure(file)

    config.playwright_browsers = playwright_browsers

    @run_in_pyodide
    def write_file(selenium, path, contents):
        path.parent.mkdir(exist_ok=True)
        import sys

        sys.path.append(str(path.parent))
        path.write_text(contents)

    write_file(selenium, Path("/test_files/test_doctest_run.py"), DOCTESTS)

    class MyPlugin:
        """Copy a couple of fixtures into the inner pytest

        If we instantiate playwright_browsers twice it breaks playwright
        If we instantiate safari twice it breaks safari
        """

        def pytest_fixture_setup(self, fixturedef, request):
            vals = {"selenium": selenium, "playwright_browsers": playwright_browsers}
            if fixturedef.argname in vals:
                val = vals[fixturedef.argname]
                cache_key = fixturedef.cache_key(request)
                fixturedef.cached_result = (val, cache_key, None)
                return val

    result = pytester.inline_run(
        file,
        "--doctest-modules",
        "--dist-dir",
        request.config.getoption("--dist-dir"),
        "--rt",
        ",".join(pytest.pyodide_runtimes),
        "--runner",
        request.config.option.runner,
        "--rootdir",
        str(file.parent),
        plugins=(MyPlugin(),),
    )
    if not pytest.pyodide_runtimes:
        result.assertoutcome(passed=1)
        return
    result.assertoutcome(passed=2, failed=1)
    result.getfailures()[0]
    captured = capsys.readouterr()
    expected = dedent(
        """
        003     >>> from js import Object # doctest: +RUN_IN_PYODIDE
        004     >>> 1 == 2
        Expected:
            True
        Got:
            False
        """
    ).strip()
    print(captured.out)
    assert expected in captured.out


def test_doctest_collect(pytester):
    BROWSERS = ["chrome", "firefox", "node", "safari"]
    BROWSER_TESTS = []
    for b in BROWSERS:
        BROWSER_TESTS.append(f"pyodide_fail[{b}]"),
        BROWSER_TESTS.append(f"pyodide_success[{b}]"),
    HOST_TESTS = ["host_success"]

    def check(rt, expected):
        path = pytester.makepyfile(DOCTESTS)
        items, reprec = pytester.inline_genitems(
            path, "--doctest-modules", "--rt", ",".join(rt)
        )
        assert sorted(
            [i.name.removeprefix("test_doctest_collect.") for i in items]
        ) == sorted(expected)

    check(BROWSERS + ["chrome-no-host"], BROWSER_TESTS)
    check(["host"], HOST_TESTS)
    check(BROWSERS + ["host"], BROWSER_TESTS + HOST_TESTS)



def test_assert(selenium):
    selenium.run_js(
        r"""
        let shouldPass;
        shouldPass = true;
        assert(() => shouldPass, "blah");
        shouldPass = false;
        let threw = false;
        try {
            assert(() => shouldPass, "blah");
        } catch(e){
            threw = true;
            if(e.message !== `Assertion failed: shouldPass\nblah`){
                throw new Error(`Unexpected message:\n${e.message}`);
            }
        }
        if(!threw){
            throw new Error("Didn't throw!");
        }
        """
    )


def test_assert_throws(selenium):
    selenium.run_js(
        r"""
        let shouldPass;
        let threw;
        assertThrows(() => { throw new TypeError("aaabbbccc") }, "TypeError", "bbc");
        assertThrows(() => { throw new TypeError("aaabbbccc") }, "TypeError", /.{3}.{3}.{3}/);
        threw = false;
        try {
            assertThrows(() => 0, "TypeError", /.*/);
        } catch(e) {
            threw = true;
            assert(() => e.message == `assertThrows(() => 0, "TypeError", /.*/) failed, no error thrown`, e.message);
        }
        assert(() => threw);
        threw = false;
        try {
            assertThrows(() => { throw new ReferenceError("blah"); }, "TypeError", /.*/);
        } catch(e) {
            threw = true;
            assert(() => e.message.endsWith("expected error of type 'TypeError' got type 'ReferenceError'"));
        }
        assert(() => threw);
        threw = false;
        try {
            assertThrows(() => { throw new TypeError("blah"); }, "TypeError", "abcd");
        } catch(e) {
            threw = true;
            console.log(`!!${e.message}!!`);
            assert(() => e.message.endsWith(`expected error message to match pattern "abcd" got:\nblah`));
        }
        assert(() => threw);
        threw = false;
        try {
            assertThrows(() => { throw new TypeError("blah"); }, "TypeError", /a..d/);
        } catch(e) {
            threw = true;
            console.log(`!!${e.message}!!`);
            assert(() => e.message.endsWith(`expected error message to match pattern /a..d/ got:\nblah`));
        }
        assert(() => threw);
        """
    )



"""
This file is not imported normally, it is loaded as a string and then exec'd
into a module called pytest_pyodide.decorator inside of Pyodide.

We use the name `pytest_pyodide.decorator` for this file for two reasons:

1. so that unpickling works smoothly
2. so that importing PyodideHandle works smoothly inside Pyodide

We could handle 1. by subclassing Unpickler and overriding find_class. Then we
could give a different name like `from pytest_pyodide.in_pyodide import
PyodideHandle` or something. But I think this current approach is the easiest
for users to make sense of. It is probably still quite confusing.

See also:
https://github.com/pyodide/pytest-pyodide/issues/43
"""

import pickle
from base64 import b64decode, b64encode
from inspect import isclass
from io import BytesIO
from typing import Any

import pyodide_js


def pointer_to_object(ptr: int) -> Any:
    """Interpret ptr as a PyObject* and convert it to the actual Python object.

    Hopefully we got our reference counting right or this will blow up!
    """
    # This was the first way I thought of to convert a pointer into a Python
    # object: use PyObject_SetItem to assign it to a dictionary.
    # ctypes doesn't seem to have an API to do this directly.
    temp: dict[int, Any] = {}
    pyodide_js._module._PyDict_SetItem(id(temp), id(0), ptr)
    return temp[0]


class PyodideHandle:
    """See documentation for the same-name class in decorator.py

    We pickle this with persistent_id (see below) so there is no need for
    __getstate__. We pickle with persistent_id because on the other side when we
    unpickle we want to inject a selenium instance so that the reference count
    can be released by the finalizer. It seems most convenient to do that
    injection with "persistent_load".
    """

    def __init__(self, obj: Any):
        self.obj = obj
        self.ptr = id(obj)


class Pickler(pickle.Pickler):
    def persistent_id(self, obj: Any) -> Any:
        if not isinstance(obj, PyodideHandle):
            return None
        pyodide_js._module._Py_IncRef(obj.ptr)
        return ("PyodideHandle", obj.ptr)

    def reducer_override(self, obj):
        try:
            from _pytest.outcomes import OutcomeException

            if isclass(obj) and issubclass(obj, OutcomeException):
                # To shorten the repr, pytest sets the __module__ of these
                # classes to builtins. This breaks pickling. Restore the correct
                # value.
                obj.__module__ = OutcomeException.__module__
        except ImportError:
            pass
        return NotImplemented


class Unpickler(pickle.Unpickler):
    def persistent_load(self, pid: Any) -> Any:
        if not isinstance(pid, tuple) or len(pid) != 2 or pid[0] != "PyodideHandle":
            raise pickle.UnpicklingError("unsupported persistent object")
        ptr = pid[1]
        # Return the actual object rather than a PyodideHandle. In practice,
        # this is much more convenient!
        return pointer_to_object(ptr)


def encode(x: Any) -> str:
    f = BytesIO()
    p = Pickler(f)
    p.dump(x)
    return b64encode(f.getvalue()).decode()


def decode(x: str) -> Any:
    return Unpickler(BytesIO(b64decode(x))).load()


async def run_in_pyodide_main(
    mod64: str, args64: str, module_filename: str, func_name: str, async_func: bool
) -> tuple[int, str]:
    """
    This actually runs the code for run_in_pyodide.
    """
    __tracebackhide__ = True

    # We've pickled and base 64 encoded the ast module and the arguments so first
    # we have to decode them.
    mod = decode(mod64)
    args: tuple[Any] = decode(args64)

    # Compile and execute the ast
    co = compile(mod, module_filename, "exec")
    d: dict[str, Any] = {}
    exec(co, d)

    try:
        # Look up the appropriate function on the module and execute it.
        # The first None fills in the "selenium" argument.
        result = d[func_name](None, *args)
        if async_func:
            result = await result
        return (0, encode(result))
    except BaseException as e:
        try:
            # If tblib is present, we can show much better tracebacks.
            from tblib import pickling_support

            def get_locals(frame):
                result = {}
                tbhide = frame.f_locals.get("__tracebackhide__")
                if tbhide:
                    result["__tracebackhide__"] = tbhide
                return result

            try:
                # works if we are using tblib >= 3.0
                pickling_support.install(get_locals=get_locals)
            except TypeError:
                # tblib < 3 or pyodide-tblib
                pickling_support.install()

        except ImportError:
            pass
        return (1, encode(e))


__all__ = ["PyodideHandle", "encode"]



import pytest

from pytest_pyodide import run_in_pyodide


@pytest.mark.xfail_browsers(
    node="Should xfail",
    firefox="Should xfail",
    chrome="Should xfail",
    safari="Should xfail",
)
@run_in_pyodide
def test_xfail_browser(selenium):
    raise AssertionError("This test should xfail")



from pathlib import Path
from textwrap import dedent

import pytest

from pytest_pyodide.hook import _filter_runtimes


def test_dist_dir(pytester):
    dist_dir = str(Path("dist").resolve())

    pytester.makepyfile(
        f"""
        import pytest
        def test_option(request):
            assert str(request.config.getoption("--dist-dir", "")) == {dist_dir!r}
        """
    )

    result = pytester.runpytest("--dist-dir", "dist")
    result.assert_outcomes(passed=1)


@pytest.mark.parametrize("runner", ["selenium", "playwright"])
def test_runner(pytester, runner):
    pytester.makepyfile(
        f"""
        import pytest
        def test_option(request):
            assert request.config.getoption("--runner") == {runner!r}
        """
    )

    result = pytester.runpytest("--runner", runner)
    result.assert_outcomes(passed=1)


def test_invalid_runner(pytester):
    runner = "blah"

    pytester.makepyfile(
        f"""
        import pytest
        def test_option(request):
            assert request.config.getoption("--runner") == {runner!r}
        """
    )

    result = pytester.runpytest("--runner", runner)

    # pytest: error argument --runner: invalid choice ...
    assert result.ret == 4


@pytest.mark.parametrize(
    "_runtime",
    [
        "invalid",
    ],
)
def test_invalid_runtime(pytester, _runtime):
    _runtime.split(",")

    pytester.makepyfile(
        """
        import pytest
        def test_option():
            assert True
        """
    )

    # TODO: catch internal errors directly?
    with pytest.raises(ValueError, match="Pytest terminal summary report not found"):
        result = pytester.runpytest("--runtime", _runtime)
        result.assert_outcomes(errors=1)


@pytest.mark.parametrize(
    "_runtime,expected",
    [
        ("chrome", (True, {"chrome"})),
        ("firefox", (True, {"firefox"})),
        ("node", (True, {"node"})),
        ("chrome,firefox", (True, {"chrome", "firefox"})),
        ("chrome-no-host,firefox", (False, {"chrome", "firefox"})),
        ("host", (True, set())),
        ("chrome-no-host,host", (True, {"chrome"})),
    ],
)
def test_filter_runtimes(_runtime, expected):
    assert _filter_runtimes(_runtime) == expected


def test_options_pytester(pytester):
    pytester.makepyfile(
        dedent(
            """
            import pytest
            from pathlib import Path

            def test_options_pytester():
                assert pytest.pyodide_run_host_test == True
                assert pytest.pyodide_runtimes == set(["chrome","firefox","safari","node"])
                assert pytest.pyodide_dist_dir == Path("some_weird_dir").resolve()
            """
        )
    )
    run_host = pytest.pyodide_run_host_test
    runtimes = pytest.pyodide_runtimes
    dist_dir = pytest.pyodide_dist_dir

    result = pytester.runpytest(
        "--dist-dir",
        Path(__file__).parents[1] / "pyodide",
        "--rt",
        "chrome,firefox,safari,node",
        "--dist-dir",
        "some_weird_dir",
    )
    result.assert_outcomes(passed=1)

    assert run_host == pytest.pyodide_run_host_test
    assert runtimes == pytest.pyodide_runtimes
    assert dist_dir == pytest.pyodide_dist_dir



import ast
from collections.abc import Callable
from copy import copy
from doctest import DocTest, DocTestRunner, register_optionflag
from pathlib import Path
from typing import cast

import pytest
from _pytest.doctest import (
    DoctestModule,
    DoctestTextfile,
    _is_doctest,
    _is_main_py,
    _is_setup_py,
)
from _pytest.python import CallSpec2
from _pytest.scope import Scope
from pytest import Collector

from . import run_in_pyodide
from .hook import ORIGINAL_MODULE_ASTS

__all__ = ["patch_doctest_runner", "collect_doctests"]


# Record the ast of this file so we can use run_in_pyodide in here
# TODO: maybe extract this as a utility function for clarity?
ORIGINAL_MODULE_ASTS[__file__] = ast.parse(
    Path(__file__).read_bytes(), filename=__file__
)
# make doctest aware of our `doctest: +RUN_IN_PYODIDE`` optionflag
RUN_IN_PYODIDE = register_optionflag("RUN_IN_PYODIDE")


def runtime_parametrize(item):
    """Force test item to be parametrized over runtimes.

    Unfortunately we have to do this manually since pytest_generate_tests only
    runs on Python tests.
    """
    scope = Scope.from_user("module", "")
    name: str
    runtimes = cast(list[str], pytest.pyodide_runtimes)
    for idx, name in enumerate(runtimes):
        newitem = copy(item)
        # dtest is the actual doctest, we have to mutate it to allow pickling so
        # better copy it too.
        newitem.dtest = copy(item.dtest)
        # Add e.g., `[chrome]` to the name, have to add it to the nodeid too or
        # else it won't be displayed on the terminal.
        newitem.name += f"[{name}]"
        newitem._nodeid += f"[{name}]"
        # Add runtime fixture to the list of fixtures and give it a specific
        # value. Normally this would be done by metafunc.parametrize but it is
        # hard to get access to that from here.
        newitem.fixturenames = ("runtime",)
        newitem.callspec = CallSpec2(
            params={"runtime": name},
            indices={"runtime": idx},
            _idlist=[name],
            _arg2scope={"runtime": scope},
        )
        yield newitem


class PyodideDoctestMixin:
    def collect(self):
        """Call super and then if test includes the RUN_IN_PYODIDE option on the
        first line, make one copy for each Pyodide runtime environment
        """
        for item in super().collect():  # type:ignore[misc]
            pyodide_test = RUN_IN_PYODIDE in item.dtest.examples[0].options
            item.dtest.pyodide_test = pyodide_test
            if not pyodide_test:
                if pytest.pyodide_run_host_test:
                    yield item
                continue
            yield from runtime_parametrize(item)


class PyodideDoctestModule(PyodideDoctestMixin, DoctestModule):
    pass


class PyodideDoctestTextfile(PyodideDoctestMixin, DoctestTextfile):
    pass


def collect_doctests(
    file_path: Path, parent: Collector, doctestmodules: bool
) -> PyodideDoctestModule | PyodideDoctestTextfile | None:
    """This is similar to _pytest.doctest.pytest_collect_file but it uses
    PyodideDoctestModule and PyodideDoctestTextfile which may run tests in
    Pyodide.
    """
    if (
        doctestmodules
        and file_path.suffix == ".py"
        and not any((_is_setup_py(file_path), _is_main_py(file_path)))
    ):
        return PyodideDoctestModule.from_parent(parent, path=file_path)
    if _is_doctest(parent.config, file_path, parent):
        return PyodideDoctestTextfile.from_parent(parent, path=file_path)
    return None


host_DocTestRunner_run = DocTestRunner.run


def patch_doctest_runner() -> None:
    DocTestRunner.run = run_doctest_in_pyodide_outer  # type: ignore[method-assign]


def run_doctest_in_pyodide_outer(
    self: DocTestRunner,
    test: DocTest,
    compileflags: int | None = None,
    out: Callable[[str], object] | None = None,
    clear_globs: bool = True,
):
    if not test.pyodide_test:  # type:ignore[attr-defined]
        # Run host test as normal
        return host_DocTestRunner_run(self, test, compileflags, out, clear_globs)

    # pytest conveniently inserts getfixture into the test globals. This saves
    # us a lot of effort.
    getfixture = test.globs["getfixture"]
    selenium = getfixture("selenium")

    # Can't pickle test with its globals. We retain the __name__ so that we can
    # figure out how to restore the globals inside of pyodide.
    test.globs = {k: test.globs[k] for k in ["__name__"]}

    # It would be nice if we could pickle DocTestRunner, but pytest has made it
    # very not pickleable. After fixing a few name resolution issues I can get
    # it to pickle successfully but I get some pickle internal error on
    # unpickling.
    #
    # So we just take the DocTestRunner apart and put it back together inside
    # Pyodide.
    optionflags = self.optionflags
    continue_on_failure = self.continue_on_failure  # type:ignore[attr-defined]

    return run_doctest_in_pyodide_inner(
        selenium, optionflags, continue_on_failure, test, compileflags, out, clear_globs
    )


@run_in_pyodide(pytest_assert_rewrites=False, packages=["pytest"])
def run_doctest_in_pyodide_inner(
    selenium,
    optionflags,
    continue_on_failure,
    test,
    compileflags,
    out,
    clear_globs,
):
    # Recreate the DocTestRunner
    #
    # It would reduce the amount of pytest internals we have to touch to
    # pickle the DocTestRunner rather than recreating it but it didn't seem
    # to work.
    from _pytest.doctest import _get_checker, _get_runner

    self = _get_runner(
        verbose=False,
        optionflags=optionflags,
        checker=_get_checker(),
        continue_on_failure=continue_on_failure,
    )

    # Put the appropriate global variables back. We do a lot less than what
    # doctest does here, but we currently don't anything but module globals
    # in our tests.
    from importlib import import_module

    try:
        mod = import_module(test.globs["__name__"])
        test.globs = mod.__dict__.copy()
    except ModuleNotFoundError:
        # Oops the test file isn't in the Pyodide file system.
        # TODO: maybe we shouldn't suppress this error??
        pass
    try:
        return self.run(test, compileflags, out, clear_globs)
    except Exception:
        # Some exceptions carry a reference to the test which cannot be
        # pickled with its global variables. Clear them out to ensure that
        # we can pickle the exception we threw.
        test.globs = {}
        raise



import contextlib
import functools
import json
from pathlib import Path


@contextlib.contextmanager
def set_webdriver_script_timeout(selenium, script_timeout: float | None):
    """Set selenium script timeout

    Parameters
    ----------
    selenum : SeleniumWrapper
       a SeleniumWrapper wrapper instance
    script_timeout : int | float
       value of the timeout in seconds
    """
    if script_timeout is not None:
        selenium.set_script_timeout(script_timeout)
    yield
    # revert to the initial value
    if script_timeout is not None:
        selenium.set_script_timeout(selenium.script_timeout)


def parse_driver_timeout(node) -> float | None:
    """Parse driver timeout value from pytest request object"""
    mark = node.get_closest_marker("driver_timeout")
    if mark is None:
        return None
    else:
        return mark.args[0]


def parse_xfail_browsers(node) -> dict[str, str]:
    mark = node.get_closest_marker("xfail_browsers")
    if mark is None:
        return {}
    return mark.kwargs


@functools.cache
def built_packages(dist_dir: Path) -> list[str]:
    """Returns the list of built package names from repodata.json"""
    repodata_path = dist_dir / "pyodide-lock.json"
    if not repodata_path.exists():
        # Try again for backwards compatibility
        repodata_path = dist_dir / "repodata.json"
    if not repodata_path.exists():
        return []
    return list(json.loads(repodata_path.read_text())["packages"].keys())


def package_is_built(package_name: str, dist_dir: Path) -> bool:
    return package_name.lower() in built_packages(dist_dir)



from pathlib import Path

import pytest

from pytest_pyodide.decorator import copy_files_to_pyodide, run_in_pyodide


@copy_files_to_pyodide(
    [(Path(__file__).parent, "non_recursive_test")], recurse_directories=False
)
def test_non_recursive_decorator(selenium):
    should_exist = f"non_recursive_test/{Path(__file__).name}"
    should_not_exist = "non_recursive_test/datafiles/in_pyodide_tests.py"
    selenium.run(
        f"""
        with open("{should_exist}") as f:
            txt = f.read()
            assert(txt.find("test_non_recursive_decorator")!=-1)
        try:
            open("{should_not_exist}","r")
            assert("File {should_not_exist} was copied")
        except IOError:
            pass
        """
    )


@copy_files_to_pyodide(
    [(Path(__file__).parent, "recursive_test")], recurse_directories=True
)
def test_recursive_decorator(selenium):
    should_exist = f"recursive_test/{Path(__file__).name}"
    should_also_exist = "recursive_test/datafiles/in_pyodide_tests.py"
    selenium.run(
        f"""
        with open("{should_exist}") as f:
            assert(txt.find("test_non_recursive_decorator")!=-1)
        try:
            open("{should_also_exist}","r")
        except IOError:
            assert("File {should_also_exist} was not copied")
        """
    )


# possible test cases = list of a)tuple, b)path, c) folder name, d) glob pattern
@copy_files_to_pyodide(
    [
        (__file__, "test.py"),
        __file__,
        (Path(__file__).parent, "testfiles"),
        (str(Path(__file__).parent) + "/test*.py", "python_only"),
    ]
)
def test_copy_files_decorator(selenium):
    auto_target_path = Path(__file__).relative_to(Path.cwd())
    selenium.run(
        f"""
    with open("test.py") as f:
        txt = f.read()
    assert(txt.find("test_copy_files_decorator test 1")!=-1)

    with open("{auto_target_path}") as f:
        txt = f.read()
    assert(txt.find("test_copy_files_decorator test 2")!=-1)

    with open("testfiles/test_copy_files.py") as f:
        txt = f.read()
    assert(txt.find("test_copy_files_decorator test 3")!=-1)

    with open("python_only/test_copy_files.py") as f:
        txt = f.read()
    assert(txt.find("test_copy_files_decorator test 4")!=-1)
    """
    )


@copy_files_to_pyodide([(__file__, "test2.py")])
@run_in_pyodide
def test_copy_files_run_in_pyodide_decorator(selenium):
    with open("test2.py") as f:
        txt = f.read()
    assert txt.find("test_copy_files_decorator test 3") != -1


def test_bad_copy_files_decorator(selenium):
    try:
        # copy all files in bad folder
        @copy_files_to_pyodide([("/*.py", "test_folder")])
        def should_throw_fn(selenium):
            pass

        should_throw_fn(selenium)
        pytest.fail(
            "Copy files to pyodide should fail if source is above current directory"
        )
    except RuntimeError:
        pass
    try:
        # copy single bad file
        @copy_files_to_pyodide([("/test.py", "test.py")])
        def should_throw_fn2(selenium):
            pass

        should_throw_fn2(selenium)
        pytest.fail(
            "Copy files to pyodide should fail if source is above current directory"
        )
    except RuntimeError:
        pass


def test_copy_files_fails_if_no_selenium():
    with pytest.raises(RuntimeError):

        @copy_files_to_pyodide([(__file__, "test2.py")])
        def should_throw_fn():
            pass

        should_throw_fn()


def test_copy_files_run_install_wheel(selenium):
    wheel_path = (
        Path(__file__).parent
    ) / "datafiles/pyodide_http-0.2.0-py3-none-any.whl"

    @copy_files_to_pyodide([(wheel_path, "pyodide_http-0.2.0-py3-none-any.whl")])
    def install_package_and_try_import(selenium):
        selenium.run("""import pyodide_http""")

    install_package_and_try_import(selenium)



from pathlib import Path

from pytest_pyodide import run_in_pyodide, spawn_web_server


@run_in_pyodide(packages=["micropip"])
async def test_install_from_pypi(selenium_standalone):
    import micropip

    await micropip.install("snowballstemmer==2.2.0")

    import snowballstemmer

    stemmer = snowballstemmer.stemmer("english")
    assert stemmer.stemWords(["university"]) == ["univers"]


def test_install_from_custom_server(selenium_standalone):
    with spawn_web_server(Path(__file__).parent / "wheels") as server:
        server_hostname, server_port, _ = server
        base_url = f"http://{server_hostname}:{server_port}/"
        url = base_url + "snowballstemmer-2.2.0-py2.py3-none-any.whl"

        selenium = selenium_standalone
        selenium.run_js(
            f"""
            await pyodide.loadPackage({url!r});
            """
        )
        selenium.run(
            """
            import snowballstemmer
            stemmer = snowballstemmer.stemmer('english')
            assert stemmer.stemWords(["university"]) == ["univers"]
            """
        )



from pathlib import Path

import pytest

from pytest_pyodide.copy_files_to_pyodide import copy_files_to_emscripten_fs
from pytest_pyodide.run_tests_inside_pyodide import run_test_in_pyodide


# fixture to copy the test file across
@pytest.fixture(scope="function")
def remote_test_file(selenium):
    datafile_path = (Path(__file__).parent / "datafiles/in_pyodide_tests.py").resolve()
    datafile_path = datafile_path.relative_to(Path.cwd())
    dest_path = Path("test_files", datafile_path)
    copy_files_to_emscripten_fs(
        [(datafile_path, dest_path)], selenium, install_wheels=False
    )
    yield dest_path


def test_fail_test(remote_test_file, selenium):
    success = run_test_in_pyodide(
        f"{remote_test_file}::test_fail", selenium, ignore_fail=True
    )
    assert success is False


@pytest.mark.xfail
def test_xfail_test(remote_test_file, selenium):
    run_test_in_pyodide(f"{remote_test_file}::test_fail", selenium, ignore_fail=False)


def test_succeed_test(remote_test_file, selenium):
    run_test_in_pyodide(
        f"{remote_test_file}::test_success", selenium, ignore_fail=False
    )


def test_running_in_pyodide(remote_test_file, selenium):
    run_test_in_pyodide(
        f"{remote_test_file}::test_check_in_pyodide", selenium, ignore_fail=False
    )


def test_pyodide_tests_skipped_inside_pyodide(remote_test_file, selenium):
    run_test_in_pyodide(
        f"{remote_test_file}::test_this_doesnt_run", selenium, ignore_fail=False
    )



import re
import sys
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from typing import ContextManager

import pytest

from .server import spawn_web_server


class ContextManagerUnwrapper:
    """Class to take a context manager (e.g. a pytest fixture or something)
    and unwrap it so that it can be used for the whole of the module.

    This is a bit of a hack, but it allows us to use some of our pytest fixtures
    without having to be inside a pytest context. Avoids significant duplication
    of the standard pytest_pyodide code here.
    """

    def __init__(self, ctx_manager: ContextManager):
        self.ctx_manager = ctx_manager
        self.value = ctx_manager.__enter__()

    def get_value(self):
        return self.value

    def __del__(self):
        self.close()

    def close(self):
        if self.ctx_manager is not None:
            self.ctx_manager.__exit__(None, None, None)
            self.value = None


@dataclass
class _SeleniumInstance:
    selenium: ContextManagerUnwrapper
    server: ContextManagerUnwrapper


_seleniums: dict[str, _SeleniumInstance] = {}
_playwright_browser_list = None
_playwright_browser_generator = None


def get_browser_pyodide(request: pytest.FixtureRequest, runtime: str):
    """Start a browser running with pyodide, ready to run pytest
    calls. If the same runtime is already running, it will
    just return that.
    """
    global _playwright_browser_generator, _playwright_browser_list
    from .fixture import _playwright_browsers, selenium_common

    if (
        request.config.option.runner.lower() == "playwright"
        and _playwright_browser_generator is None
    ):
        _playwright_browser_generator = _playwright_browsers(request)
        _playwright_browser_list = _playwright_browser_generator.__next__()
    if runtime in _seleniums:
        return _seleniums[runtime].selenium.get_value()
    web_server_main = ContextManagerUnwrapper(
        spawn_web_server(request.config.option.dist_dir)
    )
    # open pyodide
    _seleniums[runtime] = _SeleniumInstance(
        selenium=ContextManagerUnwrapper(
            selenium_common(
                request,
                runtime,
                web_server_main.get_value(),
                browsers=_playwright_browser_list,
            )
        ),
        server=web_server_main,
    )

    return _seleniums[runtime].selenium.get_value()


def _remove_pytest_capture_title(
    capture_element: ET.Element | None, title_name: str
) -> str | None:
    """
    pytest captures (even in xml) have a title line
    like ------ Capture out -------

    This helper removes that line.
    """
    if capture_element is None:
        return None
    capture_text = capture_element.text
    if not capture_text:
        return None
    lines = capture_text.splitlines()
    assert lines[0].find(" " + title_name + " ")
    ret_data = "\n".join(lines[1:])
    if re.search(r"\S", ret_data):
        return ret_data
    else:
        return None


def run_test_in_pyodide(node_tree_id, selenium, ignore_fail=False):
    """This runs a single test (identified by node_tree_id) inside
    the pyodide runtime. How it does it is by calling pytest on the
    browser pyodide with the full node ID, which is the same
    as it is locally except for the test_files folder base.

    It also has a little bit of cunning which reformats the output
    from the pyodide call to pytest, so that test failures should look
    roughly the same as they would when you are running pytest locally.
    """
    all_args = [
        node_tree_id,
        "--color=no",
        "--junitxml",
        "test_output.xml",
        "-o",
        "junit_logging=out-err",
    ]
    ret_xml = selenium.run_async(
        f"""
        import pytest
        retcode = pytest.main({all_args})

        output_xml=""
        with open("test_output.xml","r") as f:
            output_xml=f.read()
        output_xml
        """
    )
    # get the error from junitxml
    root = ET.fromstring(ret_xml)
    fails = root.findall("*/testcase[failure]")
    for fail in fails:
        stdout = fail.find("./system-out")
        stderr = fail.find("./system-err")
        failure = fail.find("failure")
        if failure is not None and failure.text:
            fail_txt = failure.text
        else:
            fail_txt = ""
        stdout_text = _remove_pytest_capture_title(stdout, "Captured Out")
        stderr_text = _remove_pytest_capture_title(stderr, "Captured Err")
        if stdout_text is not None:
            print(stdout_text)
        if stderr_text is not None:
            sys.stderr.write(stderr_text)
        if not ignore_fail:
            pytest.fail(fail_txt, pytrace=False)
        return False
    return True


def close_pyodide_browsers():
    """Close the browsers that are currently open with
    pyodide runtime initialised.

    This is done at the end of testing so that we can run more
    than one test without launching browsers each time.
    """
    global _seleniums, _playwright_browser_list, _playwright_browser_generator
    for x in _seleniums.values():
        x.selenium.close()
    _seleniums.clear()



import ast
import functools
import pickle
import sys
from base64 import b64decode, b64encode
from collections.abc import Callable, Collection
from copy import deepcopy
from io import BytesIO
from typing import Any, Protocol

import pytest

from .copy_files_to_pyodide import copy_files_to_emscripten_fs
from .hook import ORIGINAL_MODULE_ASTS, REWRITTEN_MODULE_ASTS
from .runner import _BrowserBaseRunner
from .utils import package_is_built as _package_is_built

MaybeAsyncFuncDef = ast.FunctionDef | ast.AsyncFunctionDef


def package_is_built(package_name: str):
    return _package_is_built(package_name, pytest.pyodide_dist_dir)  # type: ignore[arg-type]


class SeleniumType(Protocol):
    JavascriptException: type
    browser: str

    def load_package(self, pkgs: str | list[str]):
        ...

    def run_async(self, code: str):
        ...

    def run_js(self, code: str):
        ...


class _ReadableFileobj(Protocol):
    def read(self, __n: int) -> bytes:
        ...

    def readline(self) -> bytes:
        ...


class Unpickler(pickle.Unpickler):
    def __init__(self, file: _ReadableFileobj, selenium: SeleniumType):
        super().__init__(file)
        self.selenium = selenium

    def persistent_load(self, pid: Any) -> Any:
        if not isinstance(pid, tuple) or len(pid) != 2 or pid[0] != "PyodideHandle":
            raise pickle.UnpicklingError("unsupported persistent object")
        ptr = pid[1]
        # the PyodideHandle needs access to selenium in order to free the
        # reference count.
        return PyodideHandle(self.selenium, ptr)


class Pickler(pickle.Pickler):
    def persistent_id(self, obj: Any) -> Any:
        if not isinstance(obj, PyodideHandle):
            return None
        return ("PyodideHandle", obj.ptr)


class PyodideHandle:
    """This class allows passing a handle for a Pyodide object back to the host.

    On the host side, the handle is an opaque pointer (well we can access the
    pointer but it isn't very useful). When handed back as the argument to
    another run_in_pyodide function, it gets unpickled as the actual object.

    It's unpickled with persistent_load which injects the selenium instance.
    Because of this, we don't bother implementing __setstate__.
    """

    def __init__(self, selenium: SeleniumType, ptr: int):
        self.selenium = selenium
        self.ptr: int | None = ptr

    def __del__(self):
        if self.ptr is None:
            return
        ptr = self.ptr
        self.ptr = None
        self.selenium.run_js(
            f"""
            pyodide._module._Py_DecRef({ptr});
            """
        )


def _encode(obj: Any) -> str:
    """
    Pickle and base 64 encode obj so we can send it to Pyodide using string
    templating.
    """
    b = BytesIO()
    Pickler(b).dump(obj)
    return b64encode(b.getvalue()).decode()


def _decode(result: str, selenium: SeleniumType) -> Any:
    try:
        return Unpickler(BytesIO(b64decode(result)), selenium).load()
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(
            f"There was a problem with unpickling the return value/exception from your pyodide environment. "
            f"This usually means the type of the return value does not exist in your host environment. "
            f"The original message is: {exc}."
        ) from None


def all_args(funcdef: MaybeAsyncFuncDef) -> list[ast.arg]:
    vararg = [funcdef.args.vararg] if funcdef.args.vararg else []
    kwarg = [funcdef.args.kwarg] if funcdef.args.kwarg else []
    return (
        funcdef.args.posonlyargs
        + funcdef.args.args
        + vararg
        + funcdef.args.kwonlyargs
        + kwarg
    )


def prepare_inner_funcdef(funcdef: MaybeAsyncFuncDef) -> MaybeAsyncFuncDef:
    funcdef = deepcopy(funcdef)
    funcdef.decorator_list = []
    # Delete all type annotations
    for arg in all_args(funcdef):
        arg.annotation = None
    funcdef.returns = None
    args = funcdef.args
    args.args = all_args(funcdef)
    args.defaults = []
    args.posonlyargs = []
    args.kwonlyargs = []
    args.vararg = None
    args.kwarg = None
    funcdef.args.kw_defaults = []
    return funcdef


def value_to_name(globs: dict[str, Any], value: Any) -> ast.Name:
    """We can't put values in an `ast.Const` node unless they are deeply
    immutable (they must be of type int, float, complex, bool, string, bytes, or
    a tuple or frozenset whose entries are deeply immutable).

    So put the actual values into a dict with a unique id and return a `Name`
    node that loads the values.
    """
    index = f"v-{len(globs)}"
    globs[index] = value
    return ast.Name(index, ast.Load())


def value_to_name_or_none(
    globs: dict[str, Any], d: dict[str, Any], k: str
) -> ast.Name | None:
    if k not in d:
        return None
    return value_to_name(globs, d[k])


def prepare_outer_funcdef(
    funcdef: MaybeAsyncFuncDef, f: Callable
) -> tuple[MaybeAsyncFuncDef, dict[str, Any]]:
    funcdef = deepcopy(funcdef)
    # Clear out the decorator list.
    funcdef.decorator_list = []
    # Pull the default and annotation values off of the original function
    # object. In case they refer to local variables this gets the values from
    # the scope in which the function was originally defined. We use
    # value_to_name to stick the actual value into globs and put make a `Name`
    # node to load them out of globs.
    globs: dict[str, Any] = {}
    defaults: tuple[Any, ...] = f.__defaults__ or ()
    kwdefaults = f.__kwdefaults__ or {}
    for arg in all_args(funcdef):
        arg.annotation = value_to_name_or_none(globs, f.__annotations__, arg.arg)
    funcdef.returns = value_to_name_or_none(globs, f.__annotations__, "return")
    funcdef.args.defaults = [value_to_name(globs, x) for x in defaults]
    funcdef.args.kw_defaults = [
        value_to_name_or_none(globs, kwdefaults, arg.arg)
        for arg in funcdef.args.kwonlyargs
    ]

    return funcdef, globs


def _create_outer_func(
    run: Callable, funcdef: MaybeAsyncFuncDef, f: Callable
) -> Callable:
    """
    Create the top level item: it will be called by pytest and it calls
    run.

    If the original function looked like:

        @outer_decorators
        @run_in_pyodide
        @inner_decorators
        <async?> def func(<selenium_arg_name>, arg1, arg2, arg3):
            # do stuff

    This wrapper looks like:

        def <func_name>(<selenium_arg_name>, arg1, arg2, arg3):
            run(<selenium_arg_name>, (arg1, arg2, arg3))

    Any inner_decorators get ignored. Any outer_decorators get applied by
    the Python interpreter via the normal mechanism
    """
    funcdef, globs = prepare_outer_funcdef(funcdef, f)

    args = all_args(funcdef)
    if not args:
        raise ValueError(
            f"Function {funcdef.name} should take at least one argument whose name should start with 'selenium'"
        )

    selenium_arg_name = args[0].arg
    if not selenium_arg_name.startswith("selenium"):
        raise ValueError(
            f"Function {funcdef.name}'s first argument name '{selenium_arg_name}' should start with 'selenium'"
        )

    funcdef = ast.FunctionDef(
        name=funcdef.name,
        args=funcdef.args,
        returns=funcdef.returns,
        body=[],
        lineno=1,
        decorator_list=[],
    )

    run_id = "run-not-valid-identifier"

    # Make onwards call with two args:
    # 1. <selenium_arg_name>
    # 2. all other arguments in a tuple
    func_body = ast.parse(
        """\
        __tracebackhide__ = True; \
        return run(selenium_arg_name, (arg1, arg2, ...)) \
        """.strip()
    ).body
    onwards_call = func_body[1].value  # type: ignore[attr-defined]
    onwards_call.func = ast.Name(id=run_id, ctx=ast.Load())
    onwards_call.args[0].id = selenium_arg_name  # Set variable name
    onwards_call.args[1].elts = [  # Set tuple elements
        ast.Name(id=arg.arg, ctx=ast.Load()) for arg in args[1:]
    ]

    # Add extra <selenium_arg_name> argument
    funcdef.body = func_body
    funcdef.end_lineno = 2

    # Make a best effort to show something that isn't total nonsense in the
    # traceback for the generated function when there is an error.
    # This will show:
    # >   run(selenium_arg_name, (arg1, arg2, ...))
    # in the traceback.
    def fake_body_for_traceback(arg1, arg2, selenium_arg_name):
        run(selenium_arg_name, (arg1, arg2, ...))

    # Adjust line numbers to point into our fake function
    lineno = fake_body_for_traceback.__code__.co_firstlineno
    ast.increment_lineno(funcdef, lineno)

    mod = ast.Module([funcdef], type_ignores=[])
    ast.fix_missing_locations(mod)
    co = compile(mod, __file__, "exec")

    # Need to give our code access to the actual "run" object which it
    # invokes.
    globs.update({run_id: run})
    exec(co, globs)

    return globs[funcdef.name]


def initialize_decorator(selenium):
    from pathlib import Path

    _decorator_in_pyodide = (
        Path(__file__).parent / "_decorator_in_pyodide.py"
    ).read_text()
    selenium.run(
        f"""
def temp():
    _decorator_in_pyodide = '''{_decorator_in_pyodide}'''
    from importlib.machinery import ModuleSpec
    from importlib.util import module_from_spec

    pkgname = "pytest_pyodide"
    modname = "pytest_pyodide.decorator"
    spec = ModuleSpec(pkgname, None)
    pkg = module_from_spec(spec)
    spec = ModuleSpec(modname, None)
    mod = module_from_spec(spec)

    exec(_decorator_in_pyodide, mod.__dict__)

    import sys

    sys.modules[pkgname] = pkg
    sys.modules[modname] = mod
temp()
del temp
        """
    )


def _locate_funcdef(
    module_ast: ast.Module, f: Callable
) -> tuple[list[ast.stmt], MaybeAsyncFuncDef]:
    """Locate the statements from the original module that we need to make our
    wrapper function.

    Returns a pair:
        statements: a list of mypy magic imports that are used for mypy assertion rewrites
        funcdef: The funcdef node that makes our function
    """
    funcname = f.__name__
    func_line_no = f.__code__.co_firstlineno
    statements: list[ast.stmt] = []
    it = iter(module_ast.body)
    while True:
        try:
            node = next(it)
        except StopIteration:
            raise Exception(
                f"Didn't find function {funcname} (line {func_line_no}) in module."
            ) from None
        # We need to include the magic imports that pytest inserts
        if (
            isinstance(node, ast.Import)
            and node.names[0].asname
            and node.names[0].asname.startswith("@")
        ):
            statements.append(node)

        if (
            node.end_lineno
            and node.end_lineno > func_line_no
            and node.lineno < func_line_no
        ):
            it = iter(node.body)  # type: ignore[attr-defined]
            continue

        # We also want the function definition node
        if not isinstance(node, ast.FunctionDef | ast.AsyncFunctionDef):
            continue

        if node.lineno < func_line_no:
            continue

        if node.name != funcname:
            raise RuntimeError(
                f"Internal run_in_pyodide error: looking for function '{funcname}' but found '{node.name}'"
            )
        return (statements, node)


class run_in_pyodide:
    def __new__(cls, function: Callable | None = None, /, **kwargs):
        if function:
            # Probably we were used like:
            #
            # @run_in_pyodide
            # def f():
            #   pass
            return run_in_pyodide(**kwargs)(function)
        else:
            # Just do normal __new__ behavior
            return object.__new__(cls)

    def __init__(
        self,
        packages: Collection[str] = (),
        pytest_assert_rewrites: bool = True,
        *,
        _force_assert_rewrites: bool = False,
    ):
        """
        This decorator can be called in two ways --- with arguments and without
        arguments. If it is called without arguments, then the `function` argument
        catches the function the decorator is applied to. Otherwise, standalone and
        packages are the actual arguments to the decorator.

        See docs/testing.md for details on how to use this.

        Parameters
        ----------
        packages : List[str]
            List of packages to load before running the function in Pyodide

        pytest_assert_rewrites : bool, default = True
            If True, use pytest assertion rewrites. This gives better error messages
            when an assertion fails, but requires us to load pytest.
        """

        self._pkgs = list(packages)
        pytest_assert_rewrites = _force_assert_rewrites or (
            pytest_assert_rewrites and package_is_built("pytest")
        )

        if pytest_assert_rewrites:
            self._pkgs.append("pytest")

        self._module_asts_dict = (
            REWRITTEN_MODULE_ASTS if pytest_assert_rewrites else ORIGINAL_MODULE_ASTS
        )

        tblib_variants = (
            "pyodide-tblib",
            "tblib",
        )  # https://github.com/ionelmc/python-tblib/pull/66
        for pkg in tblib_variants:
            if package_is_built(pkg):
                self._pkgs.append(pkg)
                break

    def __call__(self, f: Callable) -> Callable:
        module = sys.modules[f.__module__]
        module_filename = module.__file__ or ""
        module_ast = self._module_asts_dict[module_filename]

        statements, funcdef = _locate_funcdef(module_ast, f)
        inner_funcdef = prepare_inner_funcdef(funcdef)
        statements.append(inner_funcdef)
        new_ast_module = ast.Module(statements, type_ignores=[])

        wrapper = _create_outer_func(self._run, funcdef, f)

        # Store information needed by self._code_template
        self._mod = new_ast_module
        self._module_filename = module_filename
        self._func_name = f.__name__
        self._async_func = isinstance(funcdef, ast.AsyncFunctionDef)
        return wrapper

    def _run(self, selenium: SeleniumType, args: tuple):
        """The main runner, called from the AST generated in _create_outer_func."""
        __tracebackhide__ = True
        code = self._code_template(args)
        if self._pkgs:
            selenium.load_package(self._pkgs)

        r = selenium.run_async(code)
        [status, result] = r

        result = _decode(result, selenium)
        if status:
            raise result
        else:
            return result

    def _code_template(self, args: tuple) -> str:
        """
        Unpickle function ast and its arguments, compile and call function, and
        if the function is async await the result. Last, if there was an
        exception, pickle it and send it back.
        """
        return f"""
        async def __tmp():
            __tracebackhide__ = True

            from pytest_pyodide.decorator import run_in_pyodide_main
            return run_in_pyodide_main(
                {_encode(self._mod)!r},
                {_encode(args)!r},
                {self._module_filename!r},
                {self._func_name!r},
                {self._async_func!r},
            )

        try:
            result = await __tmp()
        finally:
            del __tmp
        result
        """


def copy_files_to_pyodide(file_list, install_wheels=True, recurse_directories=True):
    """A decorator that copies files across to pyodide"""

    def wrap(fn):
        @functools.wraps(fn)
        def wrapped_f(*args, **argv):
            # get selenium from args
            selenium = None
            for a in args:
                if isinstance(a, _BrowserBaseRunner):
                    selenium = a
            for a in argv.values():
                if isinstance(a, _BrowserBaseRunner):
                    selenium = a
            if not selenium:
                raise RuntimeError(
                    "copy_files_to_pyodide needs a selenium argument to your test fixture"
                )
            copy_files_to_emscripten_fs(
                file_list,
                selenium,
                install_wheels=install_wheels,
                recurse_directories=recurse_directories,
            )
            return fn(*args, **argv)

        return wrapped_f

    return wrap



"""
Build test matrix for GitHub Actions
"""

import argparse
import dataclasses
import itertools
import json

DEFAULT_OS = "ubuntu-20.04, macos-11"
DEFAULT_RUNNER = "selenium, playwright"
DEFAULT_BROWSER = "chrome, firefox, node, safari, host"
DEFAULT_CHROME_VERSION = "latest"
DEFAULT_FIREFOX_VERSION = "latest"
DEFAULT_NODE_VERSION = "18, 20"
DEFAULT_PLAYWRIGHT_VERSION = (
    "1.22.0"  # TODO: versions > 1.22.0 have firefox performance issue
)


@dataclasses.dataclass
class TestConfig:
    pyodide_version: str
    os: str
    runner: str
    browser: str
    browser_version: str = ""
    playwright_version: str = ""

    @property
    def runtime(self) -> str:
        # TODO: replace all "browser" wordings to "runtime"
        return self.browser

    @property
    def runtime_version(self) -> str:
        # TODO: replace all "browser" wordings to "runtime"
        return self.browser_version


def is_valid_config(config: TestConfig) -> bool:
    """
    Check if the test-config is valid
    """

    if "macos" not in config.os and config.runtime == "safari":
        return False

    # TODO: should we support running these runtimes in macos?
    if (
        "macos" in config.os
        and config.runner == "selenium"
        and config.runtime in ("chrome", "firefox", "node")
    ):
        return False

    # It is possible to run playwright in macos, but it is not tested yet
    if "macos" in config.os and config.runner == "playwright":
        return False

    # TODO: Node is actually a runner rather than a runtime, but we treat it as a runtime conventionally...
    if config.runner == "playwright" and config.runtime == "node":
        return False

    return True


def _inject_versions_inner(
    config: TestConfig, versions: list[str], key: str
) -> list[TestConfig]:
    configs_with_versions = []
    for version in versions:
        _config = dataclasses.replace(config)
        setattr(_config, key, version)
        configs_with_versions.append(_config)

    return configs_with_versions


def inject_versions(config: TestConfig, args: dict[str, list[str]]):
    """
    Add corresponding versions to test-config
    """
    if config.runner == "playwright":
        versions = args.get("playwright_version", [])
        key = "playwright_version"
    else:
        versions = {
            "chrome": args.get("chrome_version", []),
            "firefox": args.get("firefox_version", []),
            "node": args.get("node_version", []),
            "safari": ["__IGNORED__"],  # This value will be ignored
            "host": ["__IGNORED__"],  # This value will be ignored
        }[config.runtime]
        key = "browser_version"

    return _inject_versions_inner(config, versions, key)


def remove_duplicate_configs(configs: list[TestConfig]) -> list[TestConfig]:
    """
    Remove duplicate configs.
    """
    _configs = []

    host_config = False
    for config in configs:
        if config in _configs:
            continue

        # If runtime is `host`, runner value will not be used. So we only need to keep one config
        if config.runtime == "host":
            if host_config:
                continue

            host_config = True

        _configs.append(config)

    return _configs


def build_configs(args: dict[str, list[str]]) -> list[TestConfig]:
    os = args["os"]
    pyodide_version = args["pyodide-version"]
    runner = args["runner"]
    browser = args["browser"]

    matrix = []
    for _os, _pyodide_version, _runner, _browser in itertools.product(
        os, pyodide_version, runner, browser
    ):
        config = TestConfig(_pyodide_version, _os, _runner, _browser)

        if not is_valid_config(config):
            continue

        config_with_versions = inject_versions(config, args)
        matrix.extend(config_with_versions)

    return remove_duplicate_configs(matrix)


def validate_args(args: dict[str, list[str]]):
    runners = args["runner"]
    for runner in runners:
        if runner not in ("selenium", "playwright"):
            raise ValueError(f"Invalid runner: {runner}")

    browsers = args["browser"]
    for browser in browsers:
        if browser not in ("chrome", "firefox", "node", "safari", "host"):
            raise ValueError(f"Invalid browser: {browser}")


def parse_args() -> dict[str, list[str]]:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "pyodide-version",
    )

    parser.add_argument(
        "--os",
    )
    parser.add_argument(
        "--runner",
    )
    parser.add_argument(
        "--browser",
    )
    parser.add_argument(
        "--chrome-version",
    )
    parser.add_argument(
        "--firefox-version",
    )
    parser.add_argument(
        "--node-version",
    )
    parser.add_argument(
        "--playwright-version",
    )

    defaults: dict[str, str] = {
        "os": DEFAULT_OS,
        "runner": DEFAULT_RUNNER,
        "browser": DEFAULT_BROWSER,
        "chrome_version": DEFAULT_CHROME_VERSION,
        "firefox_version": DEFAULT_FIREFOX_VERSION,
        "node_version": DEFAULT_NODE_VERSION,
        "playwright_version": DEFAULT_PLAYWRIGHT_VERSION,
    }

    args = parser.parse_args()
    args_dict: dict[str, list[str]] = {}
    for k, v in vars(args).items():
        if not v:
            v = defaults[k]

        args_dict[k] = [val.strip() for val in v.split(",")]

    validate_args(args_dict)
    return args_dict


def main():
    args = parse_args()

    configs: list[TestConfig] = build_configs(args)

    test_matrix = [dataclasses.asdict(config) for config in configs]
    print(json.dumps(test_matrix))


if __name__ == "__main__":
    main()



import pathlib

import pytest


def test_web_server_secondary(selenium, web_server_secondary):
    host, port, logs = web_server_secondary
    assert pathlib.Path(logs).exists()
    assert selenium.server_port != port


def test_host():
    assert (
        pytest.pyodide_run_host_test
    ), "this test should only run when host test is enabled"


def test_runtime(selenium):
    assert (
        pytest.pyodide_runtimes
    ), "this test should only run when runtime is specified"


def test_doctest():
    """
    >>> 1+1
    2
    """
    pass



import sys
from pathlib import Path

import pytest

sys.path.insert(0, str(Path(__file__).parent.parent / "utils"))

from build_test_matrix import (
    TestConfig,
    build_configs,
    inject_versions,
    is_valid_config,
    remove_duplicate_configs,
)


@pytest.mark.parametrize(
    "config, expected",
    [
        (TestConfig("0.21.0", "ubuntu-latest", "selenium", "firefox"), True),
        (TestConfig("0.21.0", "ubuntu-latest", "playwright", "firefox"), True),
        (TestConfig("0.21.0", "macos-latest", "selenium", "firefox"), False),
        (TestConfig("0.21.0", "macos-latest", "selenium", "chrome"), False),
        (TestConfig("0.21.0", "macos-latest", "selenium", "node"), False),
        (TestConfig("0.21.0", "macos-latest", "selenium", "safari"), True),
    ],
)
def test_build_matrix_is_valid(config, expected):
    assert is_valid_config(config) == expected


def test_build_matrix_inject_versions():
    configs = inject_versions(
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "firefox"),
        {"chrome_version": ["1.0", "2.0"], "firefox_version": ["3.0", "4.0"]},
    )

    assert len(configs) == 2
    assert configs[0].browser_version == "3.0"
    assert configs[1].browser_version == "4.0"

    configs = inject_versions(
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "chrome"),
        {"chrome_version": ["1.0", "2.0", "3.0"], "firefox_version": ["3.0", "4.0"]},
    )

    assert len(configs) == 3
    assert configs[0].browser_version == "1.0"
    assert configs[1].browser_version == "2.0"
    assert configs[2].browser_version == "3.0"

    configs = inject_versions(
        TestConfig("0.21.0", "ubuntu-latest", "playwright", "firefox"),
        {
            "chrome_version": ["1.0", "2.0"],
            "firefox_version": ["3.0", "4.0"],
            "playwright_version": ["5.0", "6.0"],
        },
    )

    assert len(configs) == 2
    assert configs[0].playwright_version == "5.0"
    assert configs[1].playwright_version == "6.0"


def test_remove_duplicate_configs():
    configs = [
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "firefox"),
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "firefox"),
    ]

    assert len(remove_duplicate_configs(configs)) == 1

    configs = [
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "firefox"),
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "chrome"),
    ]

    assert len(remove_duplicate_configs(configs)) == 2

    configs = [
        TestConfig("0.21.0", "ubuntu-latest", "selenium", "host"),
        TestConfig("0.21.0", "ubuntu-latest", "playwright", "host"),
    ]

    assert len(remove_duplicate_configs(configs)) == 1


def test_build_configs():
    configs = build_configs(
        {
            "os": ["ubuntu-latest", "macos-latest"],
            "pyodide-version": ["0.21.0"],
            "runner": ["selenium", "playwright"],
            "browser": ["firefox", "chrome", "node", "safari", "host"],
            "chrome_version": ["1.0", "2.0"],
            "firefox_version": ["3.0", "4.0"],
            "node_version": ["5.0", "6.0"],
            "playwright_version": ["7.0", "8.0"],
        }
    )

    for config in configs:
        assert is_valid_config(config)

        print(config)



from importlib.metadata import PackageNotFoundError, version

from .decorator import run_in_pyodide
from .runner import (
    NodeRunner,
    PlaywrightChromeRunner,
    PlaywrightFirefoxRunner,
    SeleniumChromeRunner,
    SeleniumFirefoxRunner,
    SeleniumSafariRunner,
)
from .server import spawn_web_server
from .utils import parse_driver_timeout, set_webdriver_script_timeout

try:
    __version__ = version("pytest-pyodide")
except PackageNotFoundError:
    # package is not installed
    pass

__all__ = [
    "NodeRunner",
    "PlaywrightChromeRunner",
    "PlaywrightFirefoxRunner",
    "SeleniumChromeRunner",
    "SeleniumFirefoxRunner",
    "SeleniumSafariRunner",
    "set_webdriver_script_timeout",
    "parse_driver_timeout",
    "run_in_pyodide",
    "spawn_web_server",
]



import pytest

from pytest_pyodide.decorator import run_in_pyodide
from pytest_pyodide.fixture import rename_fixture
from pytest_pyodide.hook import _has_standalone_fixture


@pytest.mark.parametrize("dummy", [1, 2, 3])
@run_in_pyodide
def test_selenium_standalone_refresh(selenium_standalone_refresh, dummy):
    import pathlib

    p = pathlib.Path("/hello")

    assert not p.exists()

    p.write_text("hello world")

    assert p.is_file()


def test_playwright_browsers(playwright_browsers, request):
    if request.config.option.runner.lower() != "playwright":
        pytest.skip("this test should only run when playwright is specified")

    runtimes = pytest.pyodide_runtimes

    assert set(playwright_browsers.keys()) == set(runtimes)


@run_in_pyodide
def test_jspi(selenium_jspi):
    from js import WebAssembly

    assert hasattr(WebAssembly, "Suspender")


@run_in_pyodide
def test_also_jspi(selenium_also_with_jspi):
    pass


@rename_fixture("myfixture", "myfixture_variant")
def myfunc(a, myfixture):
    return [a, myfixture]


def test_rename_fixture1():
    assert myfunc(2, 3) == [2, 3]
    assert myfunc(2, myfixture_variant=3) == [2, 3]
    assert myfunc(a=2, myfixture_variant=3) == [2, 3]


@pytest.fixture
def myfixture():
    yield 27


@pytest.fixture
def myfixture_variant():
    yield 99


@rename_fixture("myfixture", "myfixture_variant")
def test_rename_fixture2(myfixture):
    assert myfixture == 99


def test_has_standalone_fixture(pytester):
    from textwrap import dedent

    pytester.makepyfile(
        dedent(
            """
            from pytest_pyodide.fixture import rename_fixture

            @rename_fixture("selenium", "selenium_standalone")
            def test_example1(selenium):
                pass

            @rename_fixture("selenium_standalone", "selenium_standalone1")
            def test_example2(selenium_standalone):
                pass
            """
        )
    )
    node = pytester.getpathnode("test_has_standalone_fixture.py")
    r = node.collect()
    t1, t2 = r
    assert "test_example1" in t1.name
    assert "test_example2" in t2.name
    assert _has_standalone_fixture(t1)
    assert not _has_standalone_fixture(t2)



import io
import pickle
from zoneinfo import ZoneInfo

from hypothesis import HealthCheck, settings, strategies


def is_picklable(x):
    try:
        pickle.dumps(x)
        return True
    except Exception:
        return False


def is_equal_to_self(x):
    try:
        return x == x
    except Exception:
        return False


try:
    from exceptiongroup import ExceptionGroup
except ImportError:

    class ExceptionGroup:  # type: ignore[no-redef]
        pass


class NoHypothesisUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        # Only allow safe classes from builtins.
        if module == "hypothesis":
            raise pickle.UnpicklingError()
        return super().find_class(module, name)


def no_hypothesis(x):
    try:
        NoHypothesisUnpickler(io.BytesIO(pickle.dumps(x))).load()
        return True
    except Exception:
        return False


# Generate an object of any type
any_strategy = (
    strategies.from_type(type)
    .flatmap(strategies.from_type)
    .filter(lambda x: not isinstance(x, ZoneInfo))
    .filter(is_picklable)
    .filter(lambda x: not isinstance(x, ExceptionGroup))
    .filter(no_hypothesis)
)

any_equal_to_self_strategy = any_strategy.filter(is_equal_to_self)

std_hypothesis_settings = settings(
    deadline=6000,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)

strategy = (
    strategies.from_type(type)
    .flatmap(strategies.from_type)
    .filter(lambda x: not isinstance(x, ZoneInfo))
    .filter(is_picklable)
)



from collections.abc import Callable

import pytest
from hypothesis import given, settings

from pyodide.ffi import JsException
from pytest_pyodide.decorator import run_in_pyodide
from pytest_pyodide.hypothesis import any_strategy, std_hypothesis_settings
from pytest_pyodide.utils import parse_driver_timeout


@run_in_pyodide
def example_func(selenium):
    pass


@run_in_pyodide(_force_assert_rewrites=True)
def test_selenium1(selenium):
    import pytest

    with pytest.raises(AssertionError, match="assert 6 == 7"):
        x = 6
        y = 7
        assert x == y


run_in_pyodide_alias = run_in_pyodide(_force_assert_rewrites=True)


@run_in_pyodide_alias
def test_selenium2(selenium):
    import pytest

    x = 6
    y = 7
    with pytest.raises(AssertionError, match="assert 6 == 7"):
        assert x == y


@run_in_pyodide(_force_assert_rewrites=True)
async def test_selenium3(selenium):
    from asyncio import sleep

    import pytest

    await sleep(0.01)
    x = 6
    await sleep(0.01)
    y = 7
    with pytest.raises(AssertionError, match="assert 6 == 7"):
        assert x == y


def test_inner_function_closure_error(selenium):
    x = 6

    @run_in_pyodide
    def inner_function(selenium):
        assert x == 6
        return 7

    with pytest.raises(NameError, match="'x' is not defined"):
        inner_function(selenium)


def test_inner_function(selenium):
    @run_in_pyodide
    def inner_function(selenium, x):
        assert x == 6
        return 7

    assert inner_function(selenium, 6) == 7


def test_inner_function_js_exception(selenium):
    @run_in_pyodide
    def inner_function(selenium):
        from pyodide.code import run_js

        run_js("throw new Error('some error');")

    with pytest.raises(
        JsException,
        match="some error",
    ):
        inner_function(selenium)


def test_not_unpickable_return_value(selenium):
    @run_in_pyodide
    async def inner_function(selenium):
        with open("some_module.py", "w") as fp:
            fp.write("class Test: pass\n")

        from some_module import Test

        return Test()

    with pytest.raises(
        ModuleNotFoundError,
        match="There was a problem with unpickling the return.*",
    ):
        inner_function(selenium)


def complicated_decorator(attr_name: str):
    def inner_func(value):
        def dec(func):
            def wrapper(*args, **kwargs):
                wrapper.dec_info.append((attr_name, value))
                return func(*args, **kwargs)

            wrapper.dec_info = getattr(func, "dec_info", [])
            wrapper.__name__ = func.__name__
            return wrapper

        return dec

    return inner_func


d1 = complicated_decorator("testdec1")
d2 = complicated_decorator("testdec2")


@d1("a")
@d2("b")
@d1("c")
@run_in_pyodide
def example_decorator_func(selenium):
    pass


def test_selenium4(selenium_standalone):
    example_decorator_func(selenium_standalone)
    assert example_decorator_func.dec_info[-3:] == [
        ("testdec1", "a"),
        ("testdec2", "b"),
        ("testdec1", "c"),
    ]


def test_local_fail_load_package(selenium_standalone):
    selenium = selenium_standalone

    def _load_package_error(*args, **kwargs):
        raise OSError("STOP!")

    _load_package_original = selenium.load_package
    selenium.load_package = _load_package_error

    exc = None
    try:
        example_func(selenium)
    except OSError:
        exc = pytest.ExceptionInfo.from_current()
    finally:
        selenium.load_package = _load_package_original

    assert exc
    try:
        exc.getrepr()
    except IndexError as e:
        import traceback

        traceback.print_exception(e)
        raise Exception(
            "run_in_pyodide decorator badly messed up the line numbers."
            " This could crash pytest. Printed the traceback to stdout."
        ) from None


@run_in_pyodide
def test_trivial1(selenium):
    x = 7
    assert x == 7


@run_in_pyodide()
def test_trivial2(selenium):
    x = 7
    assert x == 7


@run_in_pyodide(pytest_assert_rewrites=False)
def test_trivial3(selenium):
    x = 7
    assert x == 7


@pytest.mark.parametrize("jinja2", ["jINja2", "Jinja2"])
@run_in_pyodide
def test_parametrize(selenium, jinja2):
    try:
        assert jinja2.lower() == "jinja2"
    except Exception as e:
        print(e)


@pytest.mark.skip(reason="Nope!")
@run_in_pyodide(pytest_assert_rewrites=False)
def test_skip(selenium):
    x = 6
    assert x == 7


@run_in_pyodide
async def test_run_in_pyodide_async(selenium):
    from asyncio import sleep

    x = 6
    await sleep(0.01)
    assert x == 6


@pytest.mark.skip_refcount_check
@pytest.mark.skip_pyproxy_check
@given(obj=any_strategy)
@settings(
    std_hypothesis_settings,
    max_examples=25,
)
@run_in_pyodide
def test_hypothesis(selenium_standalone, obj):
    try:
        from pyodide.ffi import to_js
    except ImportError:
        try:
            from pyodide import to_js
        except ImportError:
            raise Exception("Could not import to_js") from None

    to_js(obj)


run_in_pyodide_inner = run_in_pyodide()
run_in_pyodide_alias2 = pytest.mark.driver_timeout(40)(run_in_pyodide_inner)


@run_in_pyodide_alias2
def test_run_in_pyodide_alias(request):
    assert parse_driver_timeout(request.node) == 40


@run_in_pyodide
def set_handle(selenium, d, key, value):
    d[key] = value


@run_in_pyodide
def assert_get_handle(selenium, d, key, value):
    assert d[key] == value


@run_in_pyodide
def returns_handle(selenium):
    import __main__

    from pytest_pyodide.decorator import PyodideHandle

    o = {"a": 6}
    __main__.o = o
    handle = PyodideHandle(o)
    return handle


@run_in_pyodide
def check_refcount(selenium, n):
    import sys

    import __main__

    assert sys.getrefcount(__main__.o) == n


def test_selenium_handle(selenium):
    handle = returns_handle(selenium)
    check_refcount(selenium, 4)
    set_handle(selenium, handle, 7, 2)
    set_handle(selenium, handle, "b", 1)
    assert_get_handle(selenium, handle, "a", 6)
    assert_get_handle(selenium, handle, 7, 2)
    assert_get_handle(selenium, handle, "b", 1)
    check_refcount(selenium, 4)
    del handle
    check_refcount(selenium, 3)


def test_pytest_dot_skip(selenium):
    """Check that pytest.skip, etc will work inside @run_in_pyodide"""

    @run_in_pyodide(_force_assert_rewrites=True)
    def helper(selenium, meth):
        import pytest

        getattr(pytest, meth)("blah")

    for meth in ["skip", "fail", "xfail"]:
        with pytest.raises(getattr(pytest, meth).Exception):
            helper(selenium, meth)


def test_typehints(selenium):
    """Check that typehints are resolved correctly"""
    from collections.abc import Callable
    from typing import Any

    def fn_with_typehints(selenium, a: Any, b: Callable) -> Any:
        return [a, b]

    wrapped = run_in_pyodide(fn_with_typehints)

    assert wrapped(selenium, 5, Callable) == [5, Callable]
    assert wrapped.__annotations__ == fn_with_typehints.__annotations__


def test_posonly(selenium):
    @run_in_pyodide
    def fn_with_posonly(selenium, a, /, b):
        assert a == 5
        assert b == 7

    fn_with_posonly(selenium, 5, 7)


def test_kwonly(selenium):
    @run_in_pyodide
    def fn_with_kwonly(selenium, *, a, b):
        assert a == 5
        assert b == 7

    fn_with_kwonly(selenium, a=5, b=7)


def test_defaults1(selenium):
    """Check that defaults are kept"""

    @run_in_pyodide
    def fn_with_defaults(selenium, a=5, b=6, *, c=7, d=8):
        return [a, b, c, d]

    assert fn_with_defaults(selenium) == [5, 6, 7, 8]


def test_defaults2_from_local_variables(selenium):
    """Check that defaults that variables in defaults are resolved correctly"""
    a = Callable
    b = 6
    c = 7
    d = 8

    @run_in_pyodide
    def fn_with_defaults(selenium, a=a, b=b, *, c=c, d=d):
        return [a, b, c, d]

    assert fn_with_defaults(selenium) == [a, b, c, d]


def test_vararg(selenium):
    @run_in_pyodide
    def fn_with_vararg(selenium, a, *args: Callable):
        assert a == 5
        assert args == (6, 7, 8)

    fn_with_vararg(selenium, 5, 6, 7, 8)


def test_kwarg(selenium):
    @run_in_pyodide
    def fn_with_kwarg(selenium, a, **kwargs: Callable):
        assert a == 5
        assert kwargs == dict(b=6, c=7, d=8)

    fn_with_kwarg(selenium, a=5, b=6, c=7, d=8)


def test_complex_args(selenium):
    def fn_with_complex_args(selenium, a, /, b, *args, c, d=Callable, **kwargs):
        from collections.abc import Callable

        assert a == 5
        assert b == 6
        assert c == 7
        assert args == tuple()
        assert d == Callable
        assert kwargs == dict(a=8, q=11)

    wrapped = run_in_pyodide(fn_with_complex_args)
    import inspect

    assert inspect.signature(wrapped) == inspect.signature(fn_with_complex_args)

    wrapped(selenium, 5, b=6, a=8, c=7, q=11)

    @run_in_pyodide
    def fn_with_complex_args2(selenium, a, /, b, *args, c, d=Callable, **kwargs):
        from collections.abc import Callable

        assert a == 5
        assert b == 6
        assert c == 7
        assert args == (7, 8, 9)
        assert d == Callable
        assert kwargs == dict(a=8, q=11)

    fn_with_complex_args2(selenium, 5, 6, 7, 8, 9, a=8, c=7, q=11)
    with pytest.raises(TypeError, match="multiple values"):
        fn_with_complex_args2(selenium, 5, 6, 7, 8, 9, b=8, c=7, q=11)



import contextlib
import functools
import inspect
import os
from pathlib import Path
from typing import Any

import pytest

from .runner import (
    CHROME_FLAGS,
    FIREFOX_FLAGS,
    NodeRunner,
    PlaywrightChromeRunner,
    PlaywrightFirefoxRunner,
    SeleniumChromeRunner,
    SeleniumFirefoxRunner,
    SeleniumSafariRunner,
    _BrowserBaseRunner,
)
from .server import spawn_web_server
from .utils import parse_driver_timeout, set_webdriver_script_timeout


@pytest.fixture(scope="module")
def playwright_browsers(request):
    yield from _playwright_browsers(request)


def _playwright_browsers(request):
    if request.config.option.runner.lower() != "playwright":
        yield {}
    else:
        # import playwright here to allow running tests without playwright installation
        try:
            from playwright.sync_api import sync_playwright
        except ImportError:
            pytest.exit(
                "playwright not installed. try `pip install playwright && python -m playwright install`",
                returncode=1,
            )

        runtimes = pytest.pyodide_runtimes

        with sync_playwright() as p:
            browsers: dict[str, Any] = {}
            supported_browsers: dict[str, tuple[str, list[str]]] = {
                # browser name: (attr_name, flags)
                "firefox": ("firefox", FIREFOX_FLAGS),
                "chrome": ("chromium", CHROME_FLAGS),
                # TODO: enable webkit
                # "webkit": (),
            }
            try:
                for runtime in runtimes:  # type: ignore[attr-defined]
                    if runtime not in supported_browsers:
                        pytest.exit(
                            f"Unsupported runtime for playwright: {runtime}",
                            returncode=1,
                        )

                    attr_name, flags = supported_browsers[runtime]
                    browsers[runtime] = getattr(p, attr_name).launch(args=flags)

            except Exception as e:
                pytest.exit(f"playwright failed to launch\n{e}", returncode=1)
            try:
                yield browsers
            finally:
                for browser in browsers.values():
                    browser.close()


@contextlib.contextmanager
def selenium_common(
    request,
    runtime,
    web_server_main,
    load_pyodide=True,
    script_type="classic",
    browsers=None,
    jspi=False,
):
    """Returns an initialized selenium object.

    If `_should_skip_test` indicate that the test will be skipped,
    return None, as initializing Pyodide for selenium is expensive
    """

    server_hostname, server_port, server_log = web_server_main
    runner_type = request.config.option.runner.lower()

    runner_set: dict[tuple[str, str], type[_BrowserBaseRunner]] = {
        ("selenium", "firefox"): SeleniumFirefoxRunner,
        ("selenium", "chrome"): SeleniumChromeRunner,
        ("selenium", "safari"): SeleniumSafariRunner,
        ("selenium", "node"): NodeRunner,
        ("playwright", "firefox"): PlaywrightFirefoxRunner,
        ("playwright", "chrome"): PlaywrightChromeRunner,
        ("playwright", "node"): NodeRunner,
    }

    runner_cls = runner_set.get((runner_type, runtime))
    if runner_cls is None:
        raise AssertionError(f"Unknown runner or browser: {runner_type} / {runtime}")

    dist_dir = Path(os.getcwd(), request.config.getoption("--dist-dir"))
    runner = runner_cls(
        server_port=server_port,
        server_hostname=server_hostname,
        server_log=server_log,
        load_pyodide=load_pyodide,
        browsers=browsers,
        script_type=script_type,
        dist_dir=dist_dir,
        jspi=jspi,
    )
    try:
        yield runner
    finally:
        runner.quit()


def rename_fixture(orig_name, new_name):
    def use_variant(f):
        sig = inspect.signature(f)
        new_params = []
        for p in sig.parameters.values():
            if p.name == orig_name:
                p = p.replace(name=new_name)
            new_params.append(p)
        new_sig = sig.replace(parameters=new_params)

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if new_name in kwargs:
                kwargs[orig_name] = kwargs.pop(new_name)
            return f(*args, **kwargs)

        wrapper.__signature__ = new_sig  # type:ignore[attr-defined]
        return wrapper

    return use_variant


standalone = rename_fixture("selenium", "selenium_standalone")


@pytest.fixture(scope="function")
def selenium_standalone(request, runtime, web_server_main, playwright_browsers):
    with selenium_common(
        request, runtime, web_server_main, browsers=playwright_browsers
    ) as selenium:
        with set_webdriver_script_timeout(
            selenium, script_timeout=parse_driver_timeout(request.node)
        ):
            try:
                yield selenium
            finally:
                print(selenium.logs)


@pytest.fixture(scope="function")
def selenium_standalone_refresh(selenium):
    """
    Experimental standalone fixture which refreshes a page instead of
    instantiating a new webdriver session.
    """
    selenium.clean_logs()

    yield selenium

    selenium.refresh()
    selenium.load_pyodide()
    selenium.initialize_pyodide()
    selenium.save_state()
    selenium.restore_state()


@pytest.fixture(scope="module")
def selenium_esm(request, runtime, web_server_main, playwright_browsers):
    with selenium_common(
        request,
        runtime,
        web_server_main,
        load_pyodide=True,
        browsers=playwright_browsers,
        script_type="module",
    ) as selenium:
        with set_webdriver_script_timeout(
            selenium, script_timeout=parse_driver_timeout(request.node)
        ):
            try:
                yield selenium
            finally:
                print(selenium.logs)


@contextlib.contextmanager
def selenium_standalone_noload_common(
    request, runtime, web_server_main, playwright_browsers, script_type="classic"
):
    with selenium_common(
        request,
        runtime,
        web_server_main,
        load_pyodide=False,
        browsers=playwright_browsers,
        script_type=script_type,
    ) as selenium:
        with set_webdriver_script_timeout(
            selenium, script_timeout=parse_driver_timeout(request.node)
        ):
            try:
                yield selenium
            finally:
                print(selenium.logs)


@pytest.fixture(scope="function")
def selenium_webworker_standalone(
    request, runtime, web_server_main, playwright_browsers, script_type
):
    # Avoid loading the fixture if the test is going to be skipped
    if runtime == "firefox" and script_type == "module":
        pytest.skip("firefox does not support module type web worker")

    if runtime == "node":
        pytest.skip("no support in node")

    with selenium_standalone_noload_common(
        request, runtime, web_server_main, playwright_browsers, script_type=script_type
    ) as selenium:
        yield selenium


@pytest.fixture(scope="function")
def selenium_standalone_noload(request, runtime, web_server_main, playwright_browsers):
    """Only difference between this and selenium_webworker_standalone is that
    this also tests on node."""

    with selenium_standalone_noload_common(
        request, runtime, web_server_main, playwright_browsers
    ) as selenium:
        yield selenium


# selenium instance cached at the module level
@pytest.fixture(scope="module")
def selenium_module_scope(request, runtime, web_server_main, playwright_browsers):
    with selenium_common(
        request, runtime, web_server_main, browsers=playwright_browsers
    ) as selenium:
        yield selenium


# Hypothesis is unhappy with function scope fixtures. Instead, use the
# module scope fixture `selenium_module_scope` and use:
# `with selenium_context_manager(selenium_module_scope) as selenium`
@contextlib.contextmanager
def selenium_context_manager(selenium_module_scope):
    try:
        selenium_module_scope.clean_logs()
        yield selenium_module_scope
    finally:
        try:
            print(selenium_module_scope.logs)
        except ValueError:
            # For reasons I don't entirely understand, it is possible for
            # selenium to be closed before this is executed. In that case, just
            # skip printing the logs and we can exit cleanly.
            pass


@pytest.fixture
def selenium(request, selenium_module_scope):
    with selenium_context_manager(
        selenium_module_scope
    ) as selenium, set_webdriver_script_timeout(
        selenium, script_timeout=parse_driver_timeout(request.node)
    ):
        yield selenium


@pytest.fixture
def selenium_jspi(request, runtime, web_server_main, playwright_browsers):
    yield from selenium_jspi_inner(
        request, runtime, web_server_main, playwright_browsers
    )


def selenium_jspi_inner(request, runtime, web_server_main, playwright_browsers):
    if runtime in ["firefox", "safari"]:
        pytest.skip(f"jspi not supported in {runtime}")
    if request.config.option.runner.lower() == "playwright":
        pytest.skip("jspi not supported with playwright")
    with selenium_common(
        request, runtime, web_server_main, browsers=playwright_browsers, jspi=True
    ) as selenium, set_webdriver_script_timeout(
        selenium, script_timeout=parse_driver_timeout(request.node)
    ):
        yield selenium


@pytest.fixture(params=[False, True])
def selenium_also_with_jspi(
    selenium, request, runtime, web_server_main, playwright_browsers
):
    jspi = request.param
    if not jspi:
        yield selenium
        return
    yield from selenium_jspi_inner(
        request, runtime, web_server_main, playwright_browsers
    )


@pytest.fixture(scope="function")
def console_html_fixture(request, runtime, web_server_main, playwright_browsers):
    if runtime == "node":
        pytest.skip("no support in node")

    with selenium_common(
        request,
        runtime,
        web_server_main,
        load_pyodide=False,
        browsers=playwright_browsers,
    ) as selenium:
        selenium.goto(
            f"http://{selenium.server_hostname}:{selenium.server_port}/console.html"
        )
        selenium.javascript_setup()
        try:
            yield selenium
        finally:
            print(selenium.logs)


@pytest.fixture(scope="session")
def web_server_main(request):
    """Web server that serves files in the dist directory"""
    with spawn_web_server(request.config.option.dist_dir) as output:
        yield output


@pytest.fixture(scope="session")
def web_server_secondary(request):
    """Secondary web server that serves files dist directory"""
    with spawn_web_server(request.config.option.dist_dir) as output:
        yield output


@pytest.fixture(params=["classic", "module"], scope="module")
def script_type(request):
    return request.param



from collections.abc import MutableSequence, Sequence
from pathlib import Path
from typing import Any

from .server import spawn_web_server

_copied_files: dict[Any, MutableSequence[tuple[Path, str]]] = {}


def copy_files_to_emscripten_fs(
    file_list: Sequence[Path | str | tuple[Path | str, Path | str]],
    selenium: Any,
    install_wheels=True,
    recurse_directories=True,
):
    """
    Copies files in file_list to the emscripten file system. Files
    are passed as a list of source Path / install Path pairs.


    Parameters:
        file_list (Sequence[Path|str|tuple[Path | str, Path | str]]): A list of files, directories or glob patterns to copy as (src,destination) pairs. Destination is relative to the current
        directory on pyodide. If a single filename is passed for an entry, the destination is chosen relative to the current working directory.

        selenium : The pytest selenium fixture which hosts the pyodide to copy to.

        install_wheels (bool): If True, any wheels in the copy list are installed instead of copied.

        recurse_directories (bool): If this is True, subdirectories of directories in file_list will be copied.
    """
    if selenium not in _copied_files:
        _copied_files[selenium] = []
    new_files = []
    for list_entry in file_list:
        if isinstance(list_entry, tuple):
            src, dest_path = Path(list_entry[0]), Path(list_entry[1])
        else:
            src = Path(list_entry).resolve()
            dest_path = Path(list_entry).relative_to(Path.cwd())
        # check if it is a glob
        last_folder = ""
        last_remaining = str(src)
        glob_pattern = None
        glob_base = None
        for i, c in enumerate(str(src)):
            if c in ["*", "[", "]"]:
                glob_pattern = str(last_remaining)
                glob_base = Path(last_folder)
                print(glob_pattern)
                break
            if c == "/":
                last_folder = str(src)[: i + 1]
                last_remaining = str(src)[i + 1 :]
        if src.is_dir():
            # copy all files in directory
            if recurse_directories:
                glob_pattern = "**/*"
                glob_base = src
            else:
                glob_pattern = "./*"
                glob_base = src
        if glob_base and glob_pattern:
            # Multiple files to copy
            print(glob_base, glob_base.resolve())
            glob_base = glob_base.resolve()
            if not glob_base.is_relative_to(Path.cwd()):
                raise RuntimeError(
                    "Can only copy directories to pyodide that are below the current directory"
                )
            for f in glob_base.glob(glob_pattern):
                if f.is_dir():
                    continue
                relative_path = f.relative_to(glob_base)
                file_dest = Path(dest_path, relative_path)
                if (f, str(file_dest)) not in _copied_files[selenium]:
                    new_files.append((f, str(file_dest)))
        else:
            # Single file to copy
            src = src.resolve()
            if not src.is_relative_to(Path.cwd()):
                raise RuntimeError(
                    "Can only copy files to pyodide that are below the current directory"
                )
            if (src, str(dest_path)) not in _copied_files[selenium]:
                new_files.append((src, str(dest_path)))
    if len(new_files) == 0:
        return
    base_path = Path.cwd()
    with spawn_web_server(base_path) as server:
        server_hostname, server_port, _ = server
        base_url = f"http://{server_hostname}:{server_port}/"
        # fetch all files into the pyodide
        # n.b. this might be slow for big packages

        selenium.run(
            """
            import os
            from pathlib import Path
            from pyodide.http import pyfetch
            all_fetches = []
            all_wheels = []

            async def _fetch_file(src,dest):
                response = await pyfetch(src)
                dest.parent.mkdir(parents=True,exist_ok=True)
                with open(dest, "wb") as fp:
                    byte_data = await response.bytes()
                    fp.write(byte_data)

            """
        )
        for file, dest in new_files:
            _copied_files[selenium].append((file, dest))
            file_url = base_url + str(file.relative_to(base_path))
            if file.suffix == ".whl" and install_wheels:
                # wheel - install the wheel on the pyodide side before
                # any fetches (and don't copy it)
                selenium.run_async(
                    f"""
                    all_wheels.append("{file_url}")
                    """
                )
            else:
                # add file to fetches
                selenium.run_async(
                    f"""
                    all_fetches.append(_fetch_file("{file_url}",Path("{dest}")))
                    """
                )
        # install all wheels with micropip
        selenium.run_async(
            """
            import micropip
            await micropip.install(all_wheels)
            """
        )
        # fetch everything all at once
        selenium.run_async(
            """
            import asyncio, os, os.path
            await asyncio.gather(*all_fetches)
            """
        )



"""
This file is listed in the options.entry_points section of config.cfg so pytest
will look in here for hooks to execute.
"""

import ast
import re
import sys
from argparse import BooleanOptionalAction
from copy import copy, deepcopy
from pathlib import Path
from typing import Any, cast

import pytest
from _pytest.assertion.rewrite import AssertionRewritingHook, rewrite_asserts
from _pytest.python import (
    pytest_pycollect_makemodule as orig_pytest_pycollect_makemodule,
)
from pytest import Collector, Session

from .copy_files_to_pyodide import copy_files_to_emscripten_fs
from .run_tests_inside_pyodide import (
    close_pyodide_browsers,
    get_browser_pyodide,
    run_test_in_pyodide,
)
from .utils import parse_xfail_browsers

RUNTIMES = ["firefox", "chrome", "safari", "node"]
RUNTIMES_AND_HOST = RUNTIMES + ["host"]
RUNTIMES_NO_HOST = [f"{runtime}-no-host" for runtime in RUNTIMES]


def _filter_runtimes(runtime: str) -> tuple[bool, set[str]]:
    """Preprocess the given runtime commandline parameter

    >>> _filter_runtimes("chrome")
    (True, {'chrome'})
    >>> _filter_runtimes("chrome-no-host")
    (False, {'chrome'})
    >>> _filter_runtimes("chrome-no-host, host, firefox")
    (True, ...)
    """

    # Always run host test, unless 'no-host' is given.
    run_host = True

    # "chrome, firefox, node" ==> ["chrome", "firefox", "node"]
    runtimes = [rt.strip() for rt in runtime.split(",")]

    # remove duplicates
    runtime_set = set(runtimes)

    runtime_filtered = set()
    for rt in runtime_set:
        if rt.endswith("-no-host"):
            run_host = False
            rt = rt.replace("-no-host", "")

        runtime_filtered.add(rt)

    # If '--rt "chrome-no-host, host"' is given, we run host tests.
    run_host = run_host or ("host" in runtime_filtered)

    runtime_filtered.discard("host")

    for rt in runtime_filtered:
        if rt not in RUNTIMES:
            raise ValueError(f"Invalid runtime: {rt}")

    return run_host, runtime_filtered


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "skip_refcount_check: Don't run refcount checks",
    )

    config.addinivalue_line(
        "markers",
        "skip_pyproxy_check: Don't run pyproxy allocation checks",
    )

    config.addinivalue_line(
        "markers",
        "driver_timeout: Set script timeout in WebDriver",
    )

    config.addinivalue_line(
        "markers",
        "xfail_browsers: xfail a test in specific browsers",
    )

    config.option.dist_dir = Path(config.option.dist_dir).resolve()
    run_host, runtimes = _filter_runtimes(config.option.runtime)

    if not hasattr(pytest, "pyodide_options_stack"):
        pytest.pyodide_options_stack = []
    else:
        pytest.pyodide_options_stack.append(  # type:ignore[attr-defined]
            [
                pytest.pyodide_run_host_test,
                pytest.pyodide_runtimes,
                pytest.pyodide_dist_dir,
            ]
        )
    pytest.pyodide_run_host_test = run_host
    pytest.pyodide_runtimes = runtimes
    pytest.pyodide_dist_dir = config.option.dist_dir


def pytest_unconfigure(config):
    close_pyodide_browsers()
    try:
        (
            pytest.pyodide_run_host_test,
            pytest.pyodide_runtimes,
            pytest.pyodide_dist_dir,
        ) = pytest.pyodide_options_stack.pop()  # type:ignore[attr-defined]
    except IndexError:
        pass


@pytest.hookimpl(tryfirst=True)
def pytest_addoption(parser):
    group = parser.getgroup("general")
    group.addoption(
        "--dist-dir",
        dest="dist_dir",
        action="store",
        default="pyodide",
        help="Path to the pyodide dist directory",
        type=Path,
    )
    group.addoption(
        "--runner",
        default="selenium",
        choices=["selenium", "playwright"],
        help="Select testing frameworks, selenium or playwright (default: %(default)s)",
    )
    group.addoption(
        "--run-in-pyodide",
        action=BooleanOptionalAction,
        help="Run standard pytest tests, but in pyodide",
    )

    group.addoption(
        "--rt",
        "--runtime",
        dest="runtime",
        default="node",
        help="Select runtimes to run tests (default: %(default)s)",
    )


# We don't know the params yet, but we can set them when we do know them in
# pytest_collection
@pytest.fixture(params=[], scope="module")
def runtime(request):
    return request.param


def set_runtime_fixture_params(session):
    rt = session._fixturemanager._arg2fixturedefs["runtime"]
    rt[0].params = pytest.pyodide_runtimes


def pytest_collection(session: Session):
    from .doctest import patch_doctest_runner

    patch_doctest_runner()
    session.config.option.doctestmodules_ = session.config.option.doctestmodules
    set_runtime_fixture_params(session)


def pytest_collect_file(file_path: Path, parent: Collector):
    # Have to set doctestmodules to False to prevent original hook from
    # triggering
    parent.config.option.doctestmodules = False
    doctestmodules = getattr(parent.config.option, "doctestmodules_", False)
    from .doctest import collect_doctests

    # Call our collection hook instead. (If there are no doctests to collect,
    # collect_doctests will return None)
    return collect_doctests(file_path, parent, doctestmodules)


# Handling for pytest assertion rewrites
# First we find the pytest rewrite config. It's an attribute of the pytest
# assertion rewriting meta_path_finder, so we locate that to get the config.


def _get_pytest_rewrite_config() -> Any:
    for meta_path_finder in sys.meta_path:
        if isinstance(meta_path_finder, AssertionRewritingHook):
            break
    else:
        return None
    return meta_path_finder.config


# Now we need to parse the ast of the files, rewrite the ast, and store the
# original and rewritten ast into dictionaries. `run_in_pyodide` will look the
# ast up in the appropriate dictionary depending on whether or not it is using
# pytest assert rewrites.

REWRITE_CONFIG = _get_pytest_rewrite_config()
del _get_pytest_rewrite_config

ORIGINAL_MODULE_ASTS: dict[str, ast.Module] = {}
REWRITTEN_MODULE_ASTS: dict[str, ast.Module] = {}


def pytest_pycollect_makemodule(module_path: Path, parent: Collector) -> None:
    source = module_path.read_bytes()
    strfn = str(module_path)
    tree = ast.parse(source, filename=strfn)
    ORIGINAL_MODULE_ASTS[strfn] = tree
    tree2 = deepcopy(tree)
    rewrite_asserts(tree2, source, strfn, REWRITE_CONFIG)
    REWRITTEN_MODULE_ASTS[strfn] = tree2
    orig_pytest_pycollect_makemodule(module_path, parent)


STANDALONE_FIXTURES = [
    "selenium_standalone",
    "selenium_standalone_noload",
    "selenium_webworker_standalone",
]


def _has_standalone_fixture(item):
    for fixture in item._request.fixturenames:
        if fixture in STANDALONE_FIXTURES:
            return True

    return False


def modifyitems_run_in_pyodide(items: list[Any]):
    # TODO: get rid of this
    # if we are running tests in pyodide, then run all tests for each runtime
    new_items = []
    # if pyodide_runtimes is not a singleton this is buggy...
    # pytest_collection_modifyitems is only allowed to filter and reorder items,
    # not to add new ones...
    for runtime in pytest.pyodide_runtimes:  # type: ignore[attr-defined]
        if runtime == "host":
            continue
        for x in items:
            x = copy(x)
            x.pyodide_runtime = runtime
            new_items.append(x)
    items[:] = new_items
    return


def pytest_collection_modifyitems(items: list[Any]) -> None:
    # TODO: is this the best way to figure out if run_in_pyodide was requested?
    if items and items[0].config.option.run_in_pyodide:
        modifyitems_run_in_pyodide(items)

    # Run all Safari standalone tests first
    # Since Safari doesn't support more than one simultaneous session, we run all
    # selenium_standalone Safari tests first. We preserve the order of other
    # tests.
    OFFSET = 10000
    counter = [0]

    def _get_item_position(item):
        counter[0] += 1
        if any(
            [re.match(r"^safari[\-$]?", el) for el in item.keywords._markers.keys()]
        ) and _has_standalone_fixture(item):
            return counter[0] - OFFSET
        return counter[0]

    items[:] = sorted(items, key=_get_item_position)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    if item.config.option.run_in_pyodide:
        if not hasattr(item, "fixturenames"):
            return
        if pytest.pyodide_runtimes and "runtime" in item.fixturenames:
            pytest.skip(reason="pyodide specific test, can't run in pyodide")
        else:
            # Pass this test to pyodide runner
            # First: make sure that pyodide has the test folder copied over
            item_path = Path(item.path)
            copy_files = list(item_path.parent.rglob("*"))
            # If we have a pyodide build dist folder with wheels in, copy those over
            # and install the wheels in pyodide so we can import this package for tests
            dist_path = Path.cwd() / "dist"
            if dist_path.exists():
                copy_files.extend(list(dist_path.glob("*.whl")))

            copy_files_with_destinations = []
            for src in copy_files:
                dest = Path("test_files") / src.relative_to(Path.cwd())
                copy_files_with_destinations.append((src, dest))

            class RequestType:
                config = item.config
                node = item

            selenium = get_browser_pyodide(
                request=cast(pytest.FixtureRequest, RequestType),
                runtime=item.pyodide_runtime,
            )
            copy_files_to_emscripten_fs(
                copy_files_with_destinations, selenium, install_wheels=True
            )
    else:
        if not hasattr(item, "fixturenames"):
            # Some items like DoctestItem have no fixture
            return
        if not pytest.pyodide_runtimes and "runtime" in item.fixturenames:
            pytest.skip(reason="Non-host test")
        elif not pytest.pyodide_run_host_test and "runtime" not in item.fixturenames:
            pytest.skip("Host test")


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    if item.config.option.run_in_pyodide:

        def _run_in_pyodide(self):
            class RequestType:
                config = item.config
                node = item

            selenium = get_browser_pyodide(
                request=cast(pytest.FixtureRequest, RequestType),
                runtime=item.pyodide_runtime,
            )
            run_test_in_pyodide(self.nodeid, selenium)

        item.runtest = _run_in_pyodide.__get__(item, item.__class__)
        yield
        return

    browser = None
    for fixture in item._fixtureinfo.argnames:
        if fixture.startswith("selenium"):
            browser = item.funcargs[fixture]
            break

    if not browser:
        yield
        return

    xfail_msg = parse_xfail_browsers(item).get(browser.browser, None)
    if xfail_msg is not None:
        pytest.xfail(xfail_msg)

    yield
    return
